"""
This module can be used to solve problems related
to 2D Trusses.
"""

from cmath import inf
from sympy.core.add import Add
from sympy.core.mul import Mul
from sympy.core.symbol import Symbol
from sympy.core.sympify import sympify
from sympy import Matrix, pi
from sympy.functions.elementary.miscellaneous import sqrt
from sympy.matrices.dense import zeros
from sympy import sin, cos



class Truss:
    """
    A Truss is an assembly of members such as beams,
    connected by nodes, that create a rigid structure.
    In engineering, a truss is a structure that
    consists of two-force members only.

    Trusses are extremely important in engineering applications
    and can be seen in numerous real-world applications like bridges.

    Examples
    ========

    There is a Truss consisting of four nodes and five
    members connecting the nodes. A force P acts
    downward on the node D and there also exist pinned
    and roller joints on the nodes A and B respectively.

    .. image:: truss_example.png

    >>> from sympy.physics.continuum_mechanics.truss import Truss
    >>> t = Truss()
    >>> t.add_node("node_1", 0, 0)
    >>> t.add_node("node_2", 6, 0)
    >>> t.add_node("node_3", 2, 2)
    >>> t.add_node("node_4", 2, 0)
    >>> t.add_member("member_1", "node_1", "node_4")
    >>> t.add_member("member_2", "node_2", "node_4")
    >>> t.add_member("member_3", "node_1", "node_3")
    >>> t.add_member("member_4", "node_2", "node_3")
    >>> t.add_member("member_5", "node_3", "node_4")
    >>> t.apply_load("node_4", magnitude=10, direction=270)
    >>> t.apply_support("node_1", type="fixed")
    >>> t.apply_support("node_2", type="roller")
    """

    def __init__(self):
        """
        Initializes the class
        """
        self._nodes = []
        self._members = {}
        self._loads = {}
        self._supports = {}
        self._node_labels = []
        self._node_positions = []
        self._node_position_x = []
        self._node_position_y = []
        self._nodes_occupied = {}
        self._reaction_loads = {}
        self._internal_forces = {}
        self._node_coordinates = {}

    @property
    def nodes(self):
        """
        Returns the nodes of the truss along with their positions.
        """
        return self._nodes

    @property
    def node_labels(self):
        """
        Returns the node labels of the truss.
        """
        return self._node_labels

    @property
    def node_positions(self):
        """
        Returns the positions of the nodes of the truss.
        """
        return self._node_positions

    @property
    def members(self):
        """
        Returns the members of the truss along with the start and end points.
        """
        return self._members

    @property
    def member_labels(self):
        """
        Returns the members of the truss along with the start and end points.
        """
        return self._member_labels

    @property
    def supports(self):
        """
        Returns the nodes with provided supports along with the kind of support provided i.e.
        pinned or roller.
        """
        return self._supports

    @property
    def loads(self):
        """
        Returns the loads acting on the truss.
        """
        return self._loads

    @property
    def reaction_loads(self):
        """
        Returns the reaction forces for all supports which are all initialized to 0.
        """
        return self._reaction_loads

    @property
    def internal_forces(self):
        """
        Returns the internal forces for all members which are all initialized to 0.
        """
        return self._internal_forces

    def add_node(self, label, x, y):
        """
        This method adds a node to the truss along with its name/label and its location.

        Parameters
        ==========
        label:  String or a Symbol
            The label for a node. It is the only way to identify a particular node.

        x: Sympifyable
            The x-coordinate of the position of the node.

        y: Sympifyable
            The y-coordinate of the position of the node.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.nodes
        [('A', 0, 0)]
        >>> t.add_node('B', 3, 0)
        >>> t.nodes
        [('A', 0, 0), ('B', 3, 0)]
        """
        x = sympify(x)
        y = sympify(y)

        if label in self._node_labels:
            raise ValueError("Node needs to have a unique label")

        elif x in self._node_position_x and y in self._node_position_y and self._node_position_x.index(x)==self._node_position_y.index(y):
            raise ValueError("A node already exists at the given position")

        else :
            self._nodes.append((label, x, y))
            self._node_labels.append(label)
            self._node_positions.append((x, y))
            self._node_position_x.append(x)
            self._node_position_y.append(y)
            self._node_coordinates[label] = [x, y]

    def remove_node(self, label):
        """
        This method removes a node from the truss.

        Parameters
        ==========
        label:  String or Symbol
            The label of the node to be removed.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.nodes
        [('A', 0, 0)]
        >>> t.add_node('B', 3, 0)
        >>> t.nodes
        [('A', 0, 0), ('B', 3, 0)]
        >>> t.remove_node('A')
        >>> t.nodes
        [('B', 3, 0)]
        """
        for i in range(len(self.nodes)):
            if self._node_labels[i] == label:
                x = self._node_position_x[i]
                y = self._node_position_y[i]

        if label not in self._node_labels:
            raise ValueError("No such node exists in the truss")

        else:
            members_duplicate = self._members.copy()
            for member in members_duplicate:
                if label == self._members[member][0] or label == self._members[member][1]:
                    raise ValueError("The node given has members already attached to it")
            self._nodes.remove((label, x, y))
            self._node_labels.remove(label)
            self._node_positions.remove((x, y))
            self._node_position_x.remove(x)
            self._node_position_y.remove(y)
            if label in list(self._loads):
                self._loads.pop(label)
            if label in list(self._supports):
                self._supports.pop(label)
            self._node_coordinates.pop(label)

    def add_member(self, label, start, end):
        """
        This method adds a member between any two nodes in the given truss.

        Parameters
        ==========
        label: String or Symbol
            The label for a member. It is the only way to identify a particular member.

        start: String or Symbol
            The label of the starting point/node of the member.

        end: String or Symbol
            The label of the ending point/node of the member.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.add_node('B', 3, 0)
        >>> t.add_node('C', 2, 2)
        >>> t.add_member('AB', 'A', 'B')
        >>> t.members
        {'AB': ['A', 'B']}
        """

        if start not in self._node_labels or end not in self._node_labels or start==end:
            raise ValueError("The start and end points of the member must be unique nodes")

        elif label in list(self._members):
            raise ValueError("A member with the same label already exists for the truss")

        elif self._nodes_occupied.get((start, end)):
            raise ValueError("A member already exists between the two nodes")

        else:
            self._members[label] = [start, end]
            self._nodes_occupied[start, end] = True
            self._nodes_occupied[end, start] = True
            self._internal_forces[label] = 0

    def remove_member(self, label):
        """
        This method removes a member from the given truss.

        Parameters
        ==========
        label: String or Symbol
            The label for the member to be removed.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.add_node('B', 3, 0)
        >>> t.add_node('C', 2, 2)
        >>> t.add_member('AB', 'A', 'B')
        >>> t.add_member('AC', 'A', 'C')
        >>> t.add_member('BC', 'B', 'C')
        >>> t.members
        {'AB': ['A', 'B'], 'AC': ['A', 'C'], 'BC': ['B', 'C']}
        >>> t.remove_member('AC')
        >>> t.members
        {'AB': ['A', 'B'], 'BC': ['B', 'C']}
        """
        if label not in list(self._members):
            raise ValueError("No such member exists in the Truss")

        else:
            self._nodes_occupied.pop((self._members[label][0], self._members[label][1]))
            self._nodes_occupied.pop((self._members[label][1], self._members[label][0]))
            self._members.pop(label)
            self._internal_forces.pop(label)

    def change_node_label(self, label, new_label):
        """
        This method changes the label of a node.

        Parameters
        ==========
        label: String or Symbol
            The label of the node for which the label has
            to be changed.

        new_label: String or Symbol
            The new label of the node.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.add_node('B', 3, 0)
        >>> t.nodes
        [('A', 0, 0), ('B', 3, 0)]
        >>> t.change_node_label('A', 'C')
        >>> t.nodes
        [('C', 0, 0), ('B', 3, 0)]
        """
        if label not in self._node_labels:
            raise ValueError("No such node exists for the Truss")
        elif new_label in self._node_labels:
            raise ValueError("A node with the given label already exists")
        else:
            for node in self._nodes:
                if node[0] == label:
                    self._nodes[self._nodes.index((label, node[1], node[2]))] = (new_label, node[1], node[2])
                    self._node_labels[self._node_labels.index(node[0])] = new_label
                    self._node_coordinates[new_label] = self._node_coordinates[label]
                    self._node_coordinates.pop(label)
                    if node[0] in list(self._supports):
                        self._supports[new_label] = self._supports[node[0]]
                        self._supports.pop(node[0])
                    if new_label in list(self._supports):
                        if self._supports[new_label] == 'pinned':
                            if 'R_'+str(label)+'_x' in list(self._reaction_loads) and 'R_'+str(label)+'_y' in list(self._reaction_loads):
                                self._reaction_loads['R_'+str(new_label)+'_x'] = self._reaction_loads['R_'+str(label)+'_x']
                                self._reaction_loads['R_'+str(new_label)+'_y'] = self._reaction_loads['R_'+str(label)+'_y']
                                self._reaction_loads.pop('R_'+str(label)+'_x')
                                self._reaction_loads.pop('R_'+str(label)+'_y')
                            self._loads[new_label] = self._loads[label]
                            for load in self._loads[new_label]:
                                if load[1] == 90:
                                    load[0] -= Symbol('R_'+str(label)+'_y')
                                    if load[0] == 0:
                                        self._loads[label].remove(load)
                                    break
                            for load in self._loads[new_label]:
                                if load[1] == 0:
                                    load[0] -= Symbol('R_'+str(label)+'_x')
                                    if load[0] == 0:
                                        self._loads[label].remove(load)
                                    break
                            self.apply_load(new_label, Symbol('R_'+str(new_label)+'_x'), 0)
                            self.apply_load(new_label, Symbol('R_'+str(new_label)+'_y'), 90)
                            self._loads.pop(label)
                        elif self._supports[new_label] == 'roller':
                            self._loads[new_label] = self._loads[label]
                            for load in self._loads[label]:
                                if load[1] == 90:
                                    load[0] -= Symbol('R_'+str(label)+'_y')
                                    if load[0] == 0:
                                        self._loads[label].remove(load)
                                    break
                            self.apply_load(new_label, Symbol('R_'+str(new_label)+'_y'), 90)
                            self._loads.pop(label)
                    else:
                        if label in list(self._loads):
                            self._loads[new_label] = self._loads[label]
                            self._loads.pop(label)
                    for member in list(self._members):
                        if self._members[member][0] == node[0]:
                            self._members[member][0] = new_label
                            self._nodes_occupied[(new_label, self._members[member][1])] = True
                            self._nodes_occupied[(self._members[member][1], new_label)] = True
                            self._nodes_occupied.pop((label, self._members[member][1]))
                            self._nodes_occupied.pop((self._members[member][1], label))
                        elif self._members[member][1] == node[0]:
                            self._members[member][1] = new_label
                            self._nodes_occupied[(self._members[member][0], new_label)] = True
                            self._nodes_occupied[(new_label, self._members[member][0])] = True
                            self._nodes_occupied.pop((self._members[member][0], label))
                            self._nodes_occupied.pop((label, self._members[member][0]))

    def change_member_label(self, label, new_label):
        """
        This method changes the label of a member.

        Parameters
        ==========
        label: String or Symbol
            The label of the member for which the label has
            to be changed.

        new_label: String or Symbol
            The new label of the member.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.add_node('B', 3, 0)
        >>> t.nodes
        [('A', 0, 0), ('B', 3, 0)]
        >>> t.change_node_label('A', 'C')
        >>> t.nodes
        [('C', 0, 0), ('B', 3, 0)]
        >>> t.add_member('BC', 'B', 'C')
        >>> t.members
        {'BC': ['B', 'C']}
        >>> t.change_member_label('BC', 'BC_new')
        >>> t.members
        {'BC_new': ['B', 'C']}
        """
        if label not in list(self._members):
            raise ValueError("No such member exists for the Truss")

        else:
            members_duplicate = list(self._members).copy()
            for member in members_duplicate:
                if member == label:
                    self._members[new_label] = [self._members[member][0], self._members[member][1]]
                    self._members.pop(label)
                    self._internal_forces[new_label] = self._internal_forces[label]
                    self._internal_forces.pop(label)

    def apply_load(self, location, magnitude, direction):
        """
        This method applies an external load at a particular node

        Parameters
        ==========
        location: String or Symbol
            Label of the Node at which load is applied.

        magnitude: Sympifyable
            Magnitude of the load applied. It must always be positive and any changes in
            the direction of the load are not reflected here.

        direction: Sympifyable
            The angle, in degrees, that the load vector makes with the horizontal
            in the counter-clockwise direction. It takes the values 0 to 360,
            inclusive.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> from sympy import symbols
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.add_node('B', 3, 0)
        >>> P = symbols('P')
        >>> t.apply_load('A', P, 90)
        >>> t.apply_load('A', P/2, 45)
        >>> t.apply_load('A', P/4, 90)
        >>> t.loads
        {'A': [[P, 90], [P/2, 45], [P/4, 90]]}
        """
        magnitude = sympify(magnitude)
        direction = sympify(direction)

        if location not in self.node_labels:
            raise ValueError("Load must be applied at a known node")

        else:
            if location in list(self._loads):
                self._loads[location].append([magnitude, direction])
            else:
                self._loads[location] = [[magnitude, direction]]

    def remove_load(self, location, magnitude, direction):
        """
        This method removes an already
        present external load at a particular node

        Parameters
        ==========
        location: String or Symbol
            Label of the Node at which load is applied and is to be removed.

        magnitude: Sympifyable
            Magnitude of the load applied.

        direction: Sympifyable
            The angle, in degrees, that the load vector makes with the horizontal
            in the counter-clockwise direction. It takes the values 0 to 360,
            inclusive.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> from sympy import symbols
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.add_node('B', 3, 0)
        >>> P = symbols('P')
        >>> t.apply_load('A', P, 90)
        >>> t.apply_load('A', P/2, 45)
        >>> t.apply_load('A', P/4, 90)
        >>> t.loads
        {'A': [[P, 90], [P/2, 45], [P/4, 90]]}
        >>> t.remove_load('A', P/4, 90)
        >>> t.loads
        {'A': [[P, 90], [P/2, 45]]}
        """
        magnitude = sympify(magnitude)
        direction = sympify(direction)

        if location not in self.node_labels:
            raise ValueError("Load must be removed from a known node")

        else:
            if [magnitude, direction] not in self._loads[location]:
                raise ValueError("No load of this magnitude and direction has been applied at this node")
            else:
                self._loads[location].remove([magnitude, direction])
        if self._loads[location] == []:
            self._loads.pop(location)

    def apply_support(self, location, type):
        """
        This method adds a pinned or roller support at a particular node

        Parameters
        ==========

        location: String or Symbol
            Label of the Node at which support is added.

        type: String
            Type of the support being provided at the node.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.add_node('B', 3, 0)
        >>> t.apply_support('A', 'pinned')
        >>> t.supports
        {'A': 'pinned'}
        """
        if location not in self._node_labels:
            raise ValueError("Support must be added on a known node")

        else:
            if location not in list(self._supports):
                if type == 'pinned':
                    self.apply_load(location, Symbol('R_'+str(location)+'_x'), 0)
                    self.apply_load(location, Symbol('R_'+str(location)+'_y'), 90)
                elif type == 'roller':
                    self.apply_load(location, Symbol('R_'+str(location)+'_y'), 90)
            elif self._supports[location] == 'pinned':
                if type == 'roller':
                    self.remove_load(location, Symbol('R_'+str(location)+'_x'), 0)
            elif self._supports[location] == 'roller':
                if type == 'pinned':
                    self.apply_load(location, Symbol('R_'+str(location)+'_x'), 0)
            self._supports[location] = type

    def remove_support(self, location):
        """
        This method removes support from a particular node

        Parameters
        ==========

        location: String or Symbol
            Label of the Node at which support is to be removed.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node('A', 0, 0)
        >>> t.add_node('B', 3, 0)
        >>> t.apply_support('A', 'pinned')
        >>> t.supports
        {'A': 'pinned'}
        >>> t.remove_support('A')
        >>> t.supports
        {}
        """
        if location not in self._node_labels:
            raise ValueError("No such node exists in the Truss")

        elif location not in list(self._supports):
            raise ValueError("No support has been added to the given node")

        else:
            if self._supports[location] == 'pinned':
                self.remove_load(location, Symbol('R_'+str(location)+'_x'), 0)
                self.remove_load(location, Symbol('R_'+str(location)+'_y'), 90)
            elif self._supports[location] == 'roller':
                self.remove_load(location, Symbol('R_'+str(location)+'_y'), 90)
            self._supports.pop(location)

    def solve(self):
        """
        This method solves for all reaction forces of all supports and all internal forces
        of all the members in the truss, provided the Truss is solvable.

        A Truss is solvable if the following condition is met,

        2n >= r + m

        Where n is the number of nodes, r is the number of reaction forces, where each pinned
        support has 2 reaction forces and each roller has 1, and m is the number of members.

        The given condition is derived from the fact that a system of equations is solvable
        only when the number of variables is lesser than or equal to the number of equations.
        Equilibrium Equations in x and y directions give two equations per node giving 2n number
        equations. However, the truss needs to be stable as well and may be unstable if 2n > r + m.
        The number of variables is simply the sum of the number of reaction forces and member
        forces.

        .. note::
           The sign convention for the internal forces present in a member revolves around whether each
           force is compressive or tensile. While forming equations for each node, internal force due
           to a member on the node is assumed to be away from the node i.e. each force is assumed to
           be compressive by default. Hence, a positive value for an internal force implies the
           presence of compressive force in the member and a negative value implies a tensile force.

        Examples
        ========

        >>> from sympy.physics.continuum_mechanics.truss import Truss
        >>> t = Truss()
        >>> t.add_node("node_1", 0, 0)
        >>> t.add_node("node_2", 6, 0)
        >>> t.add_node("node_3", 2, 2)
        >>> t.add_node("node_4", 2, 0)
        >>> t.add_member("member_1", "node_1", "node_4")
        >>> t.add_member("member_2", "node_2", "node_4")
        >>> t.add_member("member_3", "node_1", "node_3")
        >>> t.add_member("member_4", "node_2", "node_3")
        >>> t.add_member("member_5", "node_3", "node_4")
        >>> t.apply_load("node_4", magnitude=10, direction=270)
        >>> t.apply_support("node_1", type="pinned")
        >>> t.apply_support("node_2", type="roller")
        >>> t.solve()
        >>> t.reaction_loads
        {'R_node_1_x': 0, 'R_node_1_y': 20/3, 'R_node_2_y': 10/3}
        >>> t.internal_forces
        {'member_1': 20/3, 'member_2': 20/3, 'member_3': -20*sqrt(2)/3, 'member_4': -10*sqrt(5)/3, 'member_5': 10}
        """
        count_reaction_loads = 0
        for node in self._nodes:
            if node[0] in list(self._supports):
                if self._supports[node[0]]=='pinned':
                    count_reaction_loads += 2
                elif self._supports[node[0]]=='roller':
                    count_reaction_loads += 1
        if 2*len(self._nodes) != len(self._members) + count_reaction_loads:
            raise ValueError("The given truss cannot be solved")
        coefficients_matrix = [[0 for i in range(2*len(self._nodes))] for j in range(2*len(self._nodes))]
        load_matrix = zeros(2*len(self.nodes), 1)
        load_matrix_row = 0
        for node in self._nodes:
            if node[0] in list(self._loads):
                for load in self._loads[node[0]]:
                    if load[0]!=Symbol('R_'+str(node[0])+'_x') and load[0]!=Symbol('R_'+str(node[0])+'_y'):
                        load_matrix[load_matrix_row] -= load[0]*cos(pi*load[1]/180)
                        load_matrix[load_matrix_row + 1] -= load[0]*sin(pi*load[1]/180)
            load_matrix_row += 2
        cols = 0
        row = 0
        for node in self._nodes:
            if node[0] in list(self._supports):
                if self._supports[node[0]]=='pinned':
                    coefficients_matrix[row][cols] += 1
                    coefficients_matrix[row+1][cols+1] += 1
                    cols += 2
                elif self._supports[node[0]]=='roller':
                    coefficients_matrix[row+1][cols] += 1
                    cols += 1
            row += 2
        for member in list(self._members):
            start = self._members[member][0]
            end = self._members[member][1]
            length = sqrt((self._node_coordinates[start][0]-self._node_coordinates[end][0])**2 + (self._node_coordinates[start][1]-self._node_coordinates[end][1])**2)
            start_index = self._node_labels.index(start)
            end_index = self._node_labels.index(end)
            horizontal_component_start = (self._node_coordinates[end][0]-self._node_coordinates[start][0])/length
            vertical_component_start = (self._node_coordinates[end][1]-self._node_coordinates[start][1])/length
            horizontal_component_end = (self._node_coordinates[start][0]-self._node_coordinates[end][0])/length
            vertical_component_end = (self._node_coordinates[start][1]-self._node_coordinates[end][1])/length
            coefficients_matrix[start_index*2][cols] += horizontal_component_start
            coefficients_matrix[start_index*2+1][cols] += vertical_component_start
            coefficients_matrix[end_index*2][cols] += horizontal_component_end
            coefficients_matrix[end_index*2+1][cols] += vertical_component_end
            cols += 1
        forces_matrix = (Matrix(coefficients_matrix)**-1)*load_matrix
        self._reaction_loads = {}
        i = 0
        min_load = inf
        for node in self._nodes:
            if node[0] in list(self._loads):
                for load in self._loads[node[0]]:
                    if type(load[0]) not in [Symbol, Mul, Add]:
                        min_load = min(min_load, load[0])
        for j in range(len(forces_matrix)):
            if type(forces_matrix[j]) not in [Symbol, Mul, Add]:
                if abs(forces_matrix[j]/min_load) <1E-10:
                    forces_matrix[j] = 0
        for node in self._nodes:
            if node[0] in list(self._supports):
                if self._supports[node[0]]=='pinned':
                    self._reaction_loads['R_'+str(node[0])+'_x'] = forces_matrix[i]
                    self._reaction_loads['R_'+str(node[0])+'_y'] = forces_matrix[i+1]
                    i += 2
                elif self._supports[node[0]]=='roller':
                    self._reaction_loads['R_'+str(node[0])+'_y'] = forces_matrix[i]
                    i += 1
        for member in list(self._members):
            self._internal_forces[member] = forces_matrix[i]
            i += 1
        return



try:
    from ctypes import c_float
except ImportError:
    pass

import pyglet.gl as pgl
from math import sqrt as _sqrt, acos as _acos


def cross(a, b):
    return (a[1] * b[2] - a[2] * b[1],
            a[2] * b[0] - a[0] * b[2],
            a[0] * b[1] - a[1] * b[0])


def dot(a, b):
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]


def mag(a):
    return _sqrt(a[0]**2 + a[1]**2 + a[2]**2)


def norm(a):
    m = mag(a)
    return (a[0] / m, a[1] / m, a[2] / m)


def get_sphere_mapping(x, y, width, height):
    x = min([max([x, 0]), width])
    y = min([max([y, 0]), height])

    sr = _sqrt((width/2)**2 + (height/2)**2)
    sx = ((x - width / 2) / sr)
    sy = ((y - height / 2) / sr)

    sz = 1.0 - sx**2 - sy**2

    if sz > 0.0:
        sz = _sqrt(sz)
        return (sx, sy, sz)
    else:
        sz = 0
        return norm((sx, sy, sz))

rad2deg = 180.0 / 3.141592


def get_spherical_rotatation(p1, p2, width, height, theta_multiplier):
    v1 = get_sphere_mapping(p1[0], p1[1], width, height)
    v2 = get_sphere_mapping(p2[0], p2[1], width, height)

    d = min(max([dot(v1, v2), -1]), 1)

    if abs(d - 1.0) < 0.000001:
        return None

    raxis = norm( cross(v1, v2) )
    rtheta = theta_multiplier * rad2deg * _acos(d)

    pgl.glPushMatrix()
    pgl.glLoadIdentity()
    pgl.glRotatef(rtheta, *raxis)
    mat = (c_float*16)()
    pgl.glGetFloatv(pgl.GL_MODELVIEW_MATRIX, mat)
    pgl.glPopMatrix()

    return mat



"""Tests for computing Galois groups. """

from sympy.abc import x
from sympy.combinatorics.galois import (
    S1TransitiveSubgroups, S2TransitiveSubgroups, S3TransitiveSubgroups,
    S4TransitiveSubgroups, S5TransitiveSubgroups, S6TransitiveSubgroups,
)
from sympy.polys.domains.rationalfield import QQ
from sympy.polys.numberfields.galoisgroups import (
    tschirnhausen_transformation,
    galois_group,
    _galois_group_degree_4_root_approx,
    _galois_group_degree_5_hybrid,
)
from sympy.polys.numberfields.subfield import field_isomorphism
from sympy.polys.polytools import Poly
from sympy.testing.pytest import raises


def test_tschirnhausen_transformation():
    for T in [
        Poly(x**2 - 2),
        Poly(x**2 + x + 1),
        Poly(x**4 + 1),
        Poly(x**4 - x**3 + x**2 - x + 1),
    ]:
        _, U = tschirnhausen_transformation(T)
        assert U.degree() == T.degree()
        assert U.is_monic
        assert U.is_irreducible
        K = QQ.alg_field_from_poly(T)
        L = QQ.alg_field_from_poly(U)
        assert field_isomorphism(K.ext, L.ext) is not None


# Test polys are from:
# Cohen, H. *A Course in Computational Algebraic Number Theory*.
test_polys_by_deg = {
    # Degree 1
    1: [
        (x, S1TransitiveSubgroups.S1, True)
    ],
    # Degree 2
    2: [
        (x**2 + x + 1, S2TransitiveSubgroups.S2, False)
    ],
    # Degree 3
    3: [
        (x**3 + x**2 - 2*x - 1, S3TransitiveSubgroups.A3, True),
        (x**3 + 2, S3TransitiveSubgroups.S3, False),
    ],
    # Degree 4
    4: [
        (x**4 + x**3 + x**2 + x + 1, S4TransitiveSubgroups.C4, False),
        (x**4 + 1, S4TransitiveSubgroups.V, True),
        (x**4 - 2, S4TransitiveSubgroups.D4, False),
        (x**4 + 8*x + 12, S4TransitiveSubgroups.A4, True),
        (x**4 + x + 1, S4TransitiveSubgroups.S4, False),
    ],
    # Degree 5
    5: [
        (x**5 + x**4 - 4*x**3 - 3*x**2 + 3*x + 1, S5TransitiveSubgroups.C5, True),
        (x**5 - 5*x + 12, S5TransitiveSubgroups.D5, True),
        (x**5 + 2, S5TransitiveSubgroups.M20, False),
        (x**5 + 20*x + 16, S5TransitiveSubgroups.A5, True),
        (x**5 - x + 1, S5TransitiveSubgroups.S5, False),
    ],
    # Degree 6
    6: [
        (x**6 + x**5 + x**4 + x**3 + x**2 + x + 1, S6TransitiveSubgroups.C6, False),
        (x**6 + 108, S6TransitiveSubgroups.S3, False),
        (x**6 + 2, S6TransitiveSubgroups.D6, False),
        (x**6 - 3*x**2 - 1, S6TransitiveSubgroups.A4, True),
        (x**6 + 3*x**3 + 3, S6TransitiveSubgroups.G18, False),
        (x**6 - 3*x**2 + 1, S6TransitiveSubgroups.A4xC2, False),
        (x**6 - 4*x**2 - 1, S6TransitiveSubgroups.S4p, True),
        (x**6 - 3*x**5 + 6*x**4 - 7*x**3 + 2*x**2 + x - 4, S6TransitiveSubgroups.S4m, False),
        (x**6 + 2*x**3 - 2, S6TransitiveSubgroups.G36m, False),
        (x**6 + 2*x**2 + 2, S6TransitiveSubgroups.S4xC2, False),
        (x**6 + 10*x**5 + 55*x**4 + 140*x**3 + 175*x**2 + 170*x + 25, S6TransitiveSubgroups.PSL2F5, True),
        (x**6 + 10*x**5 + 55*x**4 + 140*x**3 + 175*x**2 - 3019*x + 25, S6TransitiveSubgroups.PGL2F5, False),
        (x**6 + 6*x**4 + 2*x**3 + 9*x**2 + 6*x - 4, S6TransitiveSubgroups.G36p, True),
        (x**6 + 2*x**4 + 2*x**3 + x**2 + 2*x + 2, S6TransitiveSubgroups.G72, False),
        (x**6 + 24*x - 20, S6TransitiveSubgroups.A6, True),
        (x**6 + x + 1, S6TransitiveSubgroups.S6, False),
    ],
}


def test_galois_group():
    """
    Try all the test polys.
    """
    for deg in range(1, 7):
        polys = test_polys_by_deg[deg]
        for T, G, alt in polys:
            assert galois_group(T, by_name=True) == (G, alt)


def test_galois_group_degree_out_of_bounds():
    raises(ValueError, lambda: galois_group(Poly(0, x)))
    raises(ValueError, lambda: galois_group(Poly(1, x)))
    raises(ValueError, lambda: galois_group(Poly(x ** 7 + 1)))


def test_galois_group_not_by_name():
    """
    Check at least one polynomial of each supported degree, to see that
    conversion from name to group works.
    """
    for deg in range(1, 7):
        T, G_name, _ = test_polys_by_deg[deg][0]
        G, _ = galois_group(T)
        assert G == G_name.get_perm_group()


def test_galois_group_not_monic_over_ZZ():
    """
    Check that we can work with polys that are not monic over ZZ.
    """
    for deg in range(1, 7):
        T, G, alt = test_polys_by_deg[deg][0]
        assert galois_group(T/2, by_name=True) == (G, alt)


def test__galois_group_degree_4_root_approx():
    for T, G, alt in test_polys_by_deg[4]:
        assert _galois_group_degree_4_root_approx(Poly(T)) == (G, alt)


def test__galois_group_degree_5_hybrid():
    for T, G, alt in test_polys_by_deg[5]:
        assert _galois_group_degree_5_hybrid(Poly(T)) == (G, alt)


def test_AlgebraicField_galois_group():
    k = QQ.alg_field_from_poly(Poly(x**4 + 1))
    G, _ = k.galois_group(by_name=True)
    assert G == S4TransitiveSubgroups.V

    k = QQ.alg_field_from_poly(Poly(x**4 - 2))
    G, _ = k.galois_group(by_name=True)
    assert G == S4TransitiveSubgroups.D4



from sympy.core.numbers import igcd, mod_inverse
from sympy.core.power import integer_nthroot
from sympy.ntheory.residue_ntheory import _sqrt_mod_prime_power
from sympy.ntheory import isprime
from math import log, sqrt
import random

rgen = random.Random()

class SievePolynomial:
    def __init__(self, modified_coeff=(), a=None, b=None):
        """This class denotes the seive polynomial.
        If ``g(x) = (a*x + b)**2 - N``. `g(x)` can be expanded
        to ``a*x**2 + 2*a*b*x + b**2 - N``, so the coefficient
        is stored in the form `[a**2, 2*a*b, b**2 - N]`. This
        ensures faster `eval` method because we dont have to
        perform `a**2, 2*a*b, b**2` every time we call the
        `eval` method. As multiplication is more expensive
        than addition, by using modified_coefficient we get
        a faster seiving process.

        Parameters
        ==========

        modified_coeff : modified_coefficient of sieve polynomial
        a : parameter of the sieve polynomial
        b : parameter of the sieve polynomial
        """
        self.modified_coeff = modified_coeff
        self.a = a
        self.b = b

    def eval(self, x):
        """
        Compute the value of the sieve polynomial at point x.

        Parameters
        ==========

        x : Integer parameter for sieve polynomial
        """
        ans = 0
        for coeff in self.modified_coeff:
            ans *= x
            ans += coeff
        return ans


class FactorBaseElem:
    """This class stores an element of the `factor_base`.
    """
    def __init__(self, prime, tmem_p, log_p):
        """
        Initialization of factor_base_elem.

        Parameters
        ==========

        prime : prime number of the factor_base
        tmem_p : Integer square root of x**2 = n mod prime
        log_p : Compute Natural Logarithm of the prime
        """
        self.prime = prime
        self.tmem_p = tmem_p
        self.log_p = log_p
        self.soln1 = None
        self.soln2 = None
        self.a_inv = None
        self.b_ainv = None


def _generate_factor_base(prime_bound, n):
    """Generate `factor_base` for Quadratic Sieve. The `factor_base`
    consists of all the points whose ``legendre_symbol(n, p) == 1``
    and ``p < num_primes``. Along with the prime `factor_base` also stores
    natural logarithm of prime and the residue n modulo p.
    It also returns the of primes numbers in the `factor_base` which are
    close to 1000 and 5000.

    Parameters
    ==========

    prime_bound : upper prime bound of the factor_base
    n : integer to be factored
    """
    from sympy.ntheory.generate import sieve
    factor_base = []
    idx_1000, idx_5000 = None, None
    for prime in sieve.primerange(1, prime_bound):
        if pow(n, (prime - 1) // 2, prime) == 1:
            if prime > 1000 and idx_1000 is None:
                idx_1000 = len(factor_base) - 1
            if prime > 5000 and idx_5000 is None:
                idx_5000 = len(factor_base) - 1
            residue = _sqrt_mod_prime_power(n, prime, 1)[0]
            log_p = round(log(prime)*2**10)
            factor_base.append(FactorBaseElem(prime, residue, log_p))
    return idx_1000, idx_5000, factor_base


def _initialize_first_polynomial(N, M, factor_base, idx_1000, idx_5000, seed=None):
    """This step is the initialization of the 1st sieve polynomial.
    Here `a` is selected as a product of several primes of the factor_base
    such that `a` is about to ``sqrt(2*N) / M``. Other initial values of
    factor_base elem are also initialized which includes a_inv, b_ainv, soln1,
    soln2 which are used when the sieve polynomial is changed. The b_ainv
    is required for fast polynomial change as we do not have to calculate
    `2*b*mod_inverse(a, prime)` every time.
    We also ensure that the `factor_base` primes which make `a` are between
    1000 and 5000.

    Parameters
    ==========

    N : Number to be factored
    M : sieve interval
    factor_base : factor_base primes
    idx_1000 : index of prime number in the factor_base near 1000
    idx_5000 : index of prime number in the factor_base near to 5000
    seed : Generate pseudoprime numbers
    """
    if seed is not None:
        rgen.seed(seed)
    approx_val = sqrt(2*N) / M
    # `a` is a parameter of the sieve polynomial and `q` is the prime factors of `a`
    # randomly search for a combination of primes whose multiplication is close to approx_val
    # This multiplication of primes will be `a` and the primes will be `q`
    # `best_a` denotes that `a` is close to approx_val in the random search of combination
    best_a, best_q, best_ratio = None, None, None
    start = 0 if idx_1000 is None else idx_1000
    end = len(factor_base) - 1 if idx_5000 is None else idx_5000
    for _ in range(50):
        a = 1
        q = []
        while(a < approx_val):
            rand_p = 0
            while(rand_p == 0 or rand_p in q):
                rand_p = rgen.randint(start, end)
            p = factor_base[rand_p].prime
            a *= p
            q.append(rand_p)
        ratio = a / approx_val
        if best_ratio is None or abs(ratio - 1) < abs(best_ratio - 1):
            best_q = q
            best_a = a
            best_ratio = ratio

    a = best_a
    q = best_q

    B = []
    for idx, val in enumerate(q):
        q_l = factor_base[val].prime
        gamma = factor_base[val].tmem_p * mod_inverse(a // q_l, q_l) % q_l
        if gamma > q_l / 2:
            gamma = q_l - gamma
        B.append(a//q_l*gamma)

    b = sum(B)
    g = SievePolynomial([a*a, 2*a*b, b*b - N], a, b)

    for fb in factor_base:
        if a % fb.prime == 0:
            continue
        fb.a_inv = mod_inverse(a, fb.prime)
        fb.b_ainv = [2*b_elem*fb.a_inv % fb.prime for b_elem in B]
        fb.soln1 = (fb.a_inv*(fb.tmem_p - b)) % fb.prime
        fb.soln2 = (fb.a_inv*(-fb.tmem_p - b)) % fb.prime
    return g, B


def _initialize_ith_poly(N, factor_base, i, g, B):
    """Initialization stage of ith poly. After we finish sieving 1`st polynomial
    here we quickly change to the next polynomial from which we will again
    start sieving. Suppose we generated ith sieve polynomial and now we
    want to generate (i + 1)th polynomial, where ``1 <= i <= 2**(j - 1) - 1``
    where `j` is the number of prime factors of the coefficient `a`
    then this function can be used to go to the next polynomial. If
    ``i = 2**(j - 1) - 1`` then go to _initialize_first_polynomial stage.

    Parameters
    ==========

    N : number to be factored
    factor_base : factor_base primes
    i : integer denoting ith polynomial
    g : (i - 1)th polynomial
    B : array that stores a//q_l*gamma
    """
    from sympy.functions.elementary.integers import ceiling
    v = 1
    j = i
    while(j % 2 == 0):
        v += 1
        j //= 2
    if ceiling(i / (2**v)) % 2 == 1:
        neg_pow = -1
    else:
        neg_pow = 1
    b = g.b + 2*neg_pow*B[v - 1]
    a = g.a
    g = SievePolynomial([a*a, 2*a*b, b*b - N], a, b)
    for fb in factor_base:
        if a % fb.prime == 0:
            continue
        fb.soln1 = (fb.soln1 - neg_pow*fb.b_ainv[v - 1]) % fb.prime
        fb.soln2 = (fb.soln2 - neg_pow*fb.b_ainv[v - 1]) % fb.prime

    return g


def _gen_sieve_array(M, factor_base):
    """Sieve Stage of the Quadratic Sieve. For every prime in the factor_base
    that does not divide the coefficient `a` we add log_p over the sieve_array
    such that ``-M <= soln1 + i*p <=  M`` and ``-M <= soln2 + i*p <=  M`` where `i`
    is an integer. When p = 2 then log_p is only added using
    ``-M <= soln1 + i*p <=  M``.

    Parameters
    ==========

    M : sieve interval
    factor_base : factor_base primes
    """
    sieve_array = [0]*(2*M + 1)
    for factor in factor_base:
        if factor.soln1 is None: #The prime does not divides a
            continue
        for idx in range((M + factor.soln1) % factor.prime, 2*M, factor.prime):
            sieve_array[idx] += factor.log_p
        if factor.prime == 2:
            continue
        #if prime is 2 then sieve only with soln_1_p
        for idx in range((M + factor.soln2) % factor.prime, 2*M, factor.prime):
            sieve_array[idx] += factor.log_p
    return sieve_array


def _check_smoothness(num, factor_base):
    """Here we check that if `num` is a smooth number or not. If `a` is a smooth
    number then it returns a vector of prime exponents modulo 2. For example
    if a = 2 * 5**2 * 7**3 and the factor base contains {2, 3, 5, 7} then
    `a` is a smooth number and this function returns ([1, 0, 0, 1], True). If
    `a` is a partial relation which means that `a` a has one prime factor
    greater than the `factor_base` then it returns `(a, False)` which denotes `a`
    is a partial relation.

    Parameters
    ==========

    a : integer whose smootheness is to be checked
    factor_base : factor_base primes
    """
    vec = []
    if num < 0:
        vec.append(1)
        num *= -1
    else:
        vec.append(0)
     #-1 is not included in factor_base add -1 in vector
    for factor in factor_base:
        if num % factor.prime != 0:
            vec.append(0)
            continue
        factor_exp = 0
        while num % factor.prime == 0:
            factor_exp += 1
            num //= factor.prime
        vec.append(factor_exp % 2)
    if num == 1:
        return vec, True
    if isprime(num):
        return num, False
    return None, None


def _trial_division_stage(N, M, factor_base, sieve_array, sieve_poly, partial_relations, ERROR_TERM):
    """Trial division stage. Here we trial divide the values generetated
    by sieve_poly in the sieve interval and if it is a smooth number then
    it is stored in `smooth_relations`. Moreover, if we find two partial relations
    with same large prime then they are combined to form a smooth relation.
    First we iterate over sieve array and look for values which are greater
    than accumulated_val, as these values have a high chance of being smooth
    number. Then using these values we find smooth relations.
    In general, let ``t**2 = u*p modN`` and ``r**2 = v*p modN`` be two partial relations
    with the same large prime p. Then they can be combined ``(t*r/p)**2 = u*v modN``
    to form a smooth relation.

    Parameters
    ==========

    N : Number to be factored
    M : sieve interval
    factor_base : factor_base primes
    sieve_array : stores log_p values
    sieve_poly : polynomial from which we find smooth relations
    partial_relations : stores partial relations with one large prime
    ERROR_TERM : error term for accumulated_val
    """
    sqrt_n = sqrt(float(N))
    accumulated_val = log(M * sqrt_n)*2**10 - ERROR_TERM
    smooth_relations = []
    proper_factor = set()
    partial_relation_upper_bound = 128*factor_base[-1].prime
    for idx, val in enumerate(sieve_array):
        if val < accumulated_val:
            continue
        x = idx - M
        v = sieve_poly.eval(x)
        vec, is_smooth = _check_smoothness(v, factor_base)
        if is_smooth is None:#Neither smooth nor partial
            continue
        u = sieve_poly.a*x + sieve_poly.b
        # Update the partial relation
        # If 2 partial relation with same large prime is found then generate smooth relation
        if is_smooth is False:#partial relation found
            large_prime = vec
            #Consider the large_primes under 128*F
            if large_prime > partial_relation_upper_bound:
                continue
            if large_prime not in partial_relations:
                partial_relations[large_prime] = (u, v)
                continue
            else:
                u_prev, v_prev = partial_relations[large_prime]
                partial_relations.pop(large_prime)
                try:
                    large_prime_inv = mod_inverse(large_prime, N)
                except ValueError:#if large_prine divides N
                    proper_factor.add(large_prime)
                    continue
                u = u*u_prev*large_prime_inv
                v = v*v_prev // (large_prime*large_prime)
                vec, is_smooth = _check_smoothness(v, factor_base)
        #assert u*u % N == v % N
        smooth_relations.append((u, v, vec))
    return smooth_relations, proper_factor


#LINEAR ALGEBRA STAGE
def _build_matrix(smooth_relations):
    """Build a 2D matrix from smooth relations.

    Parameters
    ==========

    smooth_relations : Stores smooth relations
    """
    matrix = []
    for s_relation in smooth_relations:
        matrix.append(s_relation[2])
    return matrix


def _gauss_mod_2(A):
    """Fast gaussian reduction for modulo 2 matrix.

    Parameters
    ==========

    A : Matrix

    Examples
    ========

    >>> from sympy.ntheory.qs import _gauss_mod_2
    >>> _gauss_mod_2([[0, 1, 1], [1, 0, 1], [0, 1, 0], [1, 1, 1]])
    ([[[1, 0, 1], 3]],
     [True, True, True, False],
     [[0, 1, 0], [1, 0, 0], [0, 0, 1], [1, 0, 1]])

    Reference
    ==========

    .. [1] A fast algorithm for gaussian elimination over GF(2) and
    its implementation on the GAPP. Cetin K.Koc, Sarath N.Arachchige"""
    import copy
    matrix = copy.deepcopy(A)
    row = len(matrix)
    col = len(matrix[0])
    mark = [False]*row
    for c in range(col):
        for r in range(row):
            if matrix[r][c] == 1:
                break
        mark[r] = True
        for c1 in range(col):
            if c1 == c:
                continue
            if matrix[r][c1] == 1:
                for r2 in range(row):
                    matrix[r2][c1] = (matrix[r2][c1] + matrix[r2][c]) % 2
    dependent_row = []
    for idx, val in enumerate(mark):
        if val == False:
            dependent_row.append([matrix[idx], idx])
    return dependent_row, mark, matrix


def _find_factor(dependent_rows, mark, gauss_matrix, index, smooth_relations, N):
    """Finds proper factor of N. Here, transform the dependent rows as a
    combination of independent rows of the gauss_matrix to form the desired
    relation of the form ``X**2 = Y**2 modN``. After obtaining the desired relation
    we obtain a proper factor of N by `gcd(X - Y, N)`.

    Parameters
    ==========

    dependent_rows : denoted dependent rows in the reduced matrix form
    mark : boolean array to denoted dependent and independent rows
    gauss_matrix : Reduced form of the smooth relations matrix
    index : denoted the index of the dependent_rows
    smooth_relations : Smooth relations vectors matrix
    N : Number to be factored
    """
    idx_in_smooth = dependent_rows[index][1]
    independent_u = [smooth_relations[idx_in_smooth][0]]
    independent_v = [smooth_relations[idx_in_smooth][1]]
    dept_row = dependent_rows[index][0]

    for idx, val in enumerate(dept_row):
        if val == 1:
            for row in range(len(gauss_matrix)):
                if gauss_matrix[row][idx] == 1 and mark[row] == True:
                    independent_u.append(smooth_relations[row][0])
                    independent_v.append(smooth_relations[row][1])
                    break

    u = 1
    v = 1
    for i in independent_u:
        u *= i
    for i in independent_v:
        v *= i
    #assert u**2 % N == v % N
    v = integer_nthroot(v, 2)[0]
    return igcd(u - v, N)


def qs(N, prime_bound, M, ERROR_TERM=25, seed=1234):
    """Performs factorization using Self-Initializing Quadratic Sieve.
    In SIQS, let N be a number to be factored, and this N should not be a
    perfect power. If we find two integers such that ``X**2 = Y**2 modN`` and
    ``X != +-Y modN``, then `gcd(X + Y, N)` will reveal a proper factor of N.
    In order to find these integers X and Y we try to find relations of form
    t**2 = u modN where u is a product of small primes. If we have enough of
    these relations then we can form ``(t1*t2...ti)**2 = u1*u2...ui modN`` such that
    the right hand side is a square, thus we found a relation of ``X**2 = Y**2 modN``.

    Here, several optimizations are done like using multiple polynomials for
    sieving, fast changing between polynomials and using partial relations.
    The use of partial relations can speeds up the factoring by 2 times.

    Parameters
    ==========

    N : Number to be Factored
    prime_bound : upper bound for primes in the factor base
    M : Sieve Interval
    ERROR_TERM : Error term for checking smoothness
    threshold : Extra smooth relations for factorization
    seed : generate pseudo prime numbers

    Examples
    ========

    >>> from sympy.ntheory import qs
    >>> qs(25645121643901801, 2000, 10000)
    {5394769, 4753701529}
    >>> qs(9804659461513846513, 2000, 10000)
    {4641991, 2112166839943}

    References
    ==========

    .. [1] https://pdfs.semanticscholar.org/5c52/8a975c1405bd35c65993abf5a4edb667c1db.pdf
    .. [2] https://www.rieselprime.de/ziki/Self-initializing_quadratic_sieve
    """
    ERROR_TERM*=2**10
    rgen.seed(seed)
    idx_1000, idx_5000, factor_base = _generate_factor_base(prime_bound, N)
    smooth_relations = []
    ith_poly = 0
    partial_relations = {}
    proper_factor = set()
    threshold = 5*len(factor_base) // 100
    while True:
        if ith_poly == 0:
            ith_sieve_poly, B_array = _initialize_first_polynomial(N, M, factor_base, idx_1000, idx_5000)
        else:
            ith_sieve_poly = _initialize_ith_poly(N, factor_base, ith_poly, ith_sieve_poly, B_array)
        ith_poly += 1
        if ith_poly >= 2**(len(B_array) - 1): # time to start with a new sieve polynomial
            ith_poly = 0
        sieve_array = _gen_sieve_array(M, factor_base)
        s_rel, p_f = _trial_division_stage(N, M, factor_base, sieve_array, ith_sieve_poly, partial_relations, ERROR_TERM)
        smooth_relations += s_rel
        proper_factor |= p_f
        if len(smooth_relations) >= len(factor_base) + threshold:
            break
    matrix = _build_matrix(smooth_relations)
    dependent_row, mark, gauss_matrix = _gauss_mod_2(matrix)
    N_copy = N
    for index in range(len(dependent_row)):
        factor = _find_factor(dependent_row, mark, gauss_matrix, index, smooth_relations, N)
        if factor > 1 and factor < N:
            proper_factor.add(factor)
            while(N_copy % factor == 0):
                N_copy //= factor
            if isprime(N_copy):
                proper_factor.add(N_copy)
                break
            if(N_copy == 1):
                break
    return proper_factor



"""Simple tools for timing functions' execution, when IPython is not available. """


import timeit
import math


_scales = [1e0, 1e3, 1e6, 1e9]
_units = ['s', 'ms', '\N{GREEK SMALL LETTER MU}s', 'ns']


def timed(func, setup="pass", limit=None):
    """Adaptively measure execution time of a function. """
    timer = timeit.Timer(func, setup=setup)
    repeat, number = 3, 1

    for i in range(1, 10):
        if timer.timeit(number) >= 0.2:
            break
        elif limit is not None and number >= limit:
            break
        else:
            number *= 10

    time = min(timer.repeat(repeat, number)) / number

    if time > 0.0:
        order = min(-int(math.floor(math.log10(time)) // 3), 3)
    else:
        order = 3

    return (number, time, time*_scales[order], _units[order])


# Code for doing inline timings of recursive algorithms.

def __do_timings():
    import os
    res = os.getenv('SYMPY_TIMINGS', '')
    res = [x.strip() for x in res.split(',')]
    return set(res)

_do_timings = __do_timings()
_timestack = None


def _print_timestack(stack, level=1):
    print('-'*level, '%.2f %s%s' % (stack[2], stack[0], stack[3]))
    for s in stack[1]:
        _print_timestack(s, level + 1)


def timethis(name):
    def decorator(func):
        global _do_timings
        if name not in _do_timings:
            return func

        def wrapper(*args, **kwargs):
            from time import time
            global _timestack
            oldtimestack = _timestack
            _timestack = [func.func_name, [], 0, args]
            t1 = time()
            r = func(*args, **kwargs)
            t2 = time()
            _timestack[2] = t2 - t1
            if oldtimestack is not None:
                oldtimestack[1].append(_timestack)
                _timestack = oldtimestack
            else:
                _print_timestack(_timestack)
                _timestack = None
            return r
        return wrapper
    return decorator



""" rewrite of lambdify - This stuff is not stable at all.

It is for internal use in the new plotting module.
It may (will! see the Q'n'A in the source) be rewritten.

It's completely self contained. Especially it does not use lambdarepr.

It does not aim to replace the current lambdify. Most importantly it will never
ever support anything else than SymPy expressions (no Matrices, dictionaries
and so on).
"""


import re
from sympy.core.numbers import (I, NumberSymbol, oo, zoo)
from sympy.core.symbol import Symbol
from sympy.utilities.iterables import numbered_symbols

#  We parse the expression string into a tree that identifies functions. Then
# we translate the names of the functions and we translate also some strings
# that are not names of functions (all this according to translation
# dictionaries).
#  If the translation goes to another module (like numpy) the
# module is imported and 'func' is translated to 'module.func'.
#  If a function can not be translated, the inner nodes of that part of the
# tree are not translated. So if we have Integral(sqrt(x)), sqrt is not
# translated to np.sqrt and the Integral does not crash.
#  A namespace for all this is generated by crawling the (func, args) tree of
# the expression. The creation of this namespace involves many ugly
# workarounds.
#  The namespace consists of all the names needed for the SymPy expression and
# all the name of modules used for translation. Those modules are imported only
# as a name (import numpy as np) in order to keep the namespace small and
# manageable.

#  Please, if there is a bug, do not try to fix it here! Rewrite this by using
# the method proposed in the last Q'n'A below. That way the new function will
# work just as well, be just as simple, but it wont need any new workarounds.
#  If you insist on fixing it here, look at the workarounds in the function
# sympy_expression_namespace and in lambdify.

# Q: Why are you not using Python abstract syntax tree?
# A: Because it is more complicated and not much more powerful in this case.

# Q: What if I have Symbol('sin') or g=Function('f')?
# A: You will break the algorithm. We should use srepr to defend against this?
#  The problem with Symbol('sin') is that it will be printed as 'sin'. The
# parser will distinguish it from the function 'sin' because functions are
# detected thanks to the opening parenthesis, but the lambda expression won't
# understand the difference if we have also the sin function.
# The solution (complicated) is to use srepr and maybe ast.
#  The problem with the g=Function('f') is that it will be printed as 'f' but in
# the global namespace we have only 'g'. But as the same printer is used in the
# constructor of the namespace there will be no problem.

# Q: What if some of the printers are not printing as expected?
# A: The algorithm wont work. You must use srepr for those cases. But even
# srepr may not print well. All problems with printers should be considered
# bugs.

# Q: What about _imp_ functions?
# A: Those are taken care for by evalf. A special case treatment will work
# faster but it's not worth the code complexity.

# Q: Will ast fix all possible problems?
# A: No. You will always have to use some printer. Even srepr may not work in
# some cases. But if the printer does not work, that should be considered a
# bug.

# Q: Is there same way to fix all possible problems?
# A: Probably by constructing our strings ourself by traversing the (func,
# args) tree and creating the namespace at the same time. That actually sounds
# good.

from sympy.external import import_module
import warnings

#TODO debugging output


class vectorized_lambdify:
    """ Return a sufficiently smart, vectorized and lambdified function.

    Returns only reals.

    Explanation
    ===========

    This function uses experimental_lambdify to created a lambdified
    expression ready to be used with numpy. Many of the functions in SymPy
    are not implemented in numpy so in some cases we resort to Python cmath or
    even to evalf.

    The following translations are tried:
      only numpy complex
      - on errors raised by SymPy trying to work with ndarray:
          only Python cmath and then vectorize complex128

    When using Python cmath there is no need for evalf or float/complex
    because Python cmath calls those.

    This function never tries to mix numpy directly with evalf because numpy
    does not understand SymPy Float. If this is needed one can use the
    float_wrap_evalf/complex_wrap_evalf options of experimental_lambdify or
    better one can be explicit about the dtypes that numpy works with.
    Check numpy bug http://projects.scipy.org/numpy/ticket/1013 to know what
    types of errors to expect.
    """
    def __init__(self, args, expr):
        self.args = args
        self.expr = expr
        self.np = import_module('numpy')

        self.lambda_func_1 = experimental_lambdify(
            args, expr, use_np=True)
        self.vector_func_1 = self.lambda_func_1

        self.lambda_func_2 = experimental_lambdify(
            args, expr, use_python_cmath=True)
        self.vector_func_2 = self.np.vectorize(
            self.lambda_func_2, otypes=[complex])

        self.vector_func = self.vector_func_1
        self.failure = False

    def __call__(self, *args):
        np = self.np

        try:
            temp_args = (np.array(a, dtype=complex) for a in args)
            results = self.vector_func(*temp_args)
            results = np.ma.masked_where(
                np.abs(results.imag) > 1e-7 * np.abs(results),
                results.real, copy=False)
            return results
        except ValueError:
            if self.failure:
                raise

            self.failure = True
            self.vector_func = self.vector_func_2
            warnings.warn(
                'The evaluation of the expression is problematic. '
                'We are trying a failback method that may still work. '
                'Please report this as a bug.')
            return self.__call__(*args)


class lambdify:
    """Returns the lambdified function.

    Explanation
    ===========

    This function uses experimental_lambdify to create a lambdified
    expression. It uses cmath to lambdify the expression. If the function
    is not implemented in Python cmath, Python cmath calls evalf on those
    functions.
    """

    def __init__(self, args, expr):
        self.args = args
        self.expr = expr
        self.lambda_func_1 = experimental_lambdify(
            args, expr, use_python_cmath=True, use_evalf=True)
        self.lambda_func_2 = experimental_lambdify(
            args, expr, use_python_math=True, use_evalf=True)
        self.lambda_func_3 = experimental_lambdify(
            args, expr, use_evalf=True, complex_wrap_evalf=True)
        self.lambda_func = self.lambda_func_1
        self.failure = False

    def __call__(self, args):
        try:
            #The result can be sympy.Float. Hence wrap it with complex type.
            result = complex(self.lambda_func(args))
            if abs(result.imag) > 1e-7 * abs(result):
                return None
            return result.real
        except (ZeroDivisionError, OverflowError):
            return None
        except TypeError as e:
            if self.failure:
                raise e

            if self.lambda_func == self.lambda_func_1:
                self.lambda_func = self.lambda_func_2
                return self.__call__(args)

            self.failure = True
            self.lambda_func = self.lambda_func_3
            warnings.warn(
                'The evaluation of the expression is problematic. '
                'We are trying a failback method that may still work. '
                'Please report this as a bug.', stacklevel=2)
            return self.__call__(args)


def experimental_lambdify(*args, **kwargs):
    l = Lambdifier(*args, **kwargs)
    return l


class Lambdifier:
    def __init__(self, args, expr, print_lambda=False, use_evalf=False,
                 float_wrap_evalf=False, complex_wrap_evalf=False,
                 use_np=False, use_python_math=False, use_python_cmath=False,
                 use_interval=False):

        self.print_lambda = print_lambda
        self.use_evalf = use_evalf
        self.float_wrap_evalf = float_wrap_evalf
        self.complex_wrap_evalf = complex_wrap_evalf
        self.use_np = use_np
        self.use_python_math = use_python_math
        self.use_python_cmath = use_python_cmath
        self.use_interval = use_interval

        # Constructing the argument string
        # - check
        if not all(isinstance(a, Symbol) for a in args):
            raise ValueError('The arguments must be Symbols.')
        # - use numbered symbols
        syms = numbered_symbols(exclude=expr.free_symbols)
        newargs = [next(syms) for _ in args]
        expr = expr.xreplace(dict(zip(args, newargs)))
        argstr = ', '.join([str(a) for a in newargs])
        del syms, newargs, args

        # Constructing the translation dictionaries and making the translation
        self.dict_str = self.get_dict_str()
        self.dict_fun = self.get_dict_fun()
        exprstr = str(expr)
        newexpr = self.tree2str_translate(self.str2tree(exprstr))

        # Constructing the namespaces
        namespace = {}
        namespace.update(self.sympy_atoms_namespace(expr))
        namespace.update(self.sympy_expression_namespace(expr))
        # XXX Workaround
        # Ugly workaround because Pow(a,Half) prints as sqrt(a)
        # and sympy_expression_namespace can not catch it.
        from sympy.functions.elementary.miscellaneous import sqrt
        namespace.update({'sqrt': sqrt})
        namespace.update({'Eq': lambda x, y: x == y})
        namespace.update({'Ne': lambda x, y: x != y})
        # End workaround.
        if use_python_math:
            namespace.update({'math': __import__('math')})
        if use_python_cmath:
            namespace.update({'cmath': __import__('cmath')})
        if use_np:
            try:
                namespace.update({'np': __import__('numpy')})
            except ImportError:
                raise ImportError(
                    'experimental_lambdify failed to import numpy.')
        if use_interval:
            namespace.update({'imath': __import__(
                'sympy.plotting.intervalmath', fromlist=['intervalmath'])})
            namespace.update({'math': __import__('math')})

        # Construct the lambda
        if self.print_lambda:
            print(newexpr)
        eval_str = 'lambda %s : ( %s )' % (argstr, newexpr)
        self.eval_str = eval_str
        exec("MYNEWLAMBDA = %s" % eval_str, namespace)
        self.lambda_func = namespace['MYNEWLAMBDA']

    def __call__(self, *args, **kwargs):
        return self.lambda_func(*args, **kwargs)


    ##############################################################################
    # Dicts for translating from SymPy to other modules
    ##############################################################################
    ###
    # builtins
    ###
    # Functions with different names in builtins
    builtin_functions_different = {
        'Min': 'min',
        'Max': 'max',
        'Abs': 'abs',
    }

    # Strings that should be translated
    builtin_not_functions = {
        'I': '1j',
#        'oo': '1e400',
    }

    ###
    # numpy
    ###

    # Functions that are the same in numpy
    numpy_functions_same = [
        'sin', 'cos', 'tan', 'sinh', 'cosh', 'tanh', 'exp', 'log',
        'sqrt', 'floor', 'conjugate',
    ]

    # Functions with different names in numpy
    numpy_functions_different = {
        "acos": "arccos",
        "acosh": "arccosh",
        "arg": "angle",
        "asin": "arcsin",
        "asinh": "arcsinh",
        "atan": "arctan",
        "atan2": "arctan2",
        "atanh": "arctanh",
        "ceiling": "ceil",
        "im": "imag",
        "ln": "log",
        "Max": "amax",
        "Min": "amin",
        "re": "real",
        "Abs": "abs",
    }

    # Strings that should be translated
    numpy_not_functions = {
        'pi': 'np.pi',
        'oo': 'np.inf',
        'E': 'np.e',
    }

    ###
    # Python math
    ###

    # Functions that are the same in math
    math_functions_same = [
        'sin', 'cos', 'tan', 'asin', 'acos', 'atan', 'atan2',
        'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh',
        'exp', 'log', 'erf', 'sqrt', 'floor', 'factorial', 'gamma',
    ]

    # Functions with different names in math
    math_functions_different = {
        'ceiling': 'ceil',
        'ln': 'log',
        'loggamma': 'lgamma'
    }

    # Strings that should be translated
    math_not_functions = {
        'pi': 'math.pi',
        'E': 'math.e',
    }

    ###
    # Python cmath
    ###

    # Functions that are the same in cmath
    cmath_functions_same = [
        'sin', 'cos', 'tan', 'asin', 'acos', 'atan',
        'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh',
        'exp', 'log', 'sqrt',
    ]

    # Functions with different names in cmath
    cmath_functions_different = {
        'ln': 'log',
        'arg': 'phase',
    }

    # Strings that should be translated
    cmath_not_functions = {
        'pi': 'cmath.pi',
        'E': 'cmath.e',
    }

    ###
    # intervalmath
    ###

    interval_not_functions = {
        'pi': 'math.pi',
        'E': 'math.e'
    }

    interval_functions_same = [
        'sin', 'cos', 'exp', 'tan', 'atan', 'log',
        'sqrt', 'cosh', 'sinh', 'tanh', 'floor',
        'acos', 'asin', 'acosh', 'asinh', 'atanh',
        'Abs', 'And', 'Or'
    ]

    interval_functions_different = {
        'Min': 'imin',
        'Max': 'imax',
        'ceiling': 'ceil',

    }

    ###
    # mpmath, etc
    ###
    #TODO

    ###
    # Create the final ordered tuples of dictionaries
    ###

    # For strings
    def get_dict_str(self):
        dict_str = dict(self.builtin_not_functions)
        if self.use_np:
            dict_str.update(self.numpy_not_functions)
        if self.use_python_math:
            dict_str.update(self.math_not_functions)
        if self.use_python_cmath:
            dict_str.update(self.cmath_not_functions)
        if self.use_interval:
            dict_str.update(self.interval_not_functions)
        return dict_str

    # For functions
    def get_dict_fun(self):
        dict_fun = dict(self.builtin_functions_different)
        if self.use_np:
            for s in self.numpy_functions_same:
                dict_fun[s] = 'np.' + s
            for k, v in self.numpy_functions_different.items():
                dict_fun[k] = 'np.' + v
        if self.use_python_math:
            for s in self.math_functions_same:
                dict_fun[s] = 'math.' + s
            for k, v in self.math_functions_different.items():
                dict_fun[k] = 'math.' + v
        if self.use_python_cmath:
            for s in self.cmath_functions_same:
                dict_fun[s] = 'cmath.' + s
            for k, v in self.cmath_functions_different.items():
                dict_fun[k] = 'cmath.' + v
        if self.use_interval:
            for s in self.interval_functions_same:
                dict_fun[s] = 'imath.' + s
            for k, v in self.interval_functions_different.items():
                dict_fun[k] = 'imath.' + v
        return dict_fun

    ##############################################################################
    # The translator functions, tree parsers, etc.
    ##############################################################################

    def str2tree(self, exprstr):
        """Converts an expression string to a tree.

        Explanation
        ===========

        Functions are represented by ('func_name(', tree_of_arguments).
        Other expressions are (head_string, mid_tree, tail_str).
        Expressions that do not contain functions are directly returned.

        Examples
        ========

        >>> from sympy.abc import x, y, z
        >>> from sympy import Integral, sin
        >>> from sympy.plotting.experimental_lambdify import Lambdifier
        >>> str2tree = Lambdifier([x], x).str2tree

        >>> str2tree(str(Integral(x, (x, 1, y))))
        ('', ('Integral(', 'x, (x, 1, y)'), ')')
        >>> str2tree(str(x+y))
        'x + y'
        >>> str2tree(str(x+y*sin(z)+1))
        ('x + y*', ('sin(', 'z'), ') + 1')
        >>> str2tree('sin(y*(y + 1.1) + (sin(y)))')
        ('', ('sin(', ('y*(y + 1.1) + (', ('sin(', 'y'), '))')), ')')
        """
        #matches the first 'function_name('
        first_par = re.search(r'(\w+\()', exprstr)
        if first_par is None:
            return exprstr
        else:
            start = first_par.start()
            end = first_par.end()
            head = exprstr[:start]
            func = exprstr[start:end]
            tail = exprstr[end:]
            count = 0
            for i, c in enumerate(tail):
                if c == '(':
                    count += 1
                elif c == ')':
                    count -= 1
                if count == -1:
                    break
            func_tail = self.str2tree(tail[:i])
            tail = self.str2tree(tail[i:])
            return (head, (func, func_tail), tail)

    @classmethod
    def tree2str(cls, tree):
        """Converts a tree to string without translations.

        Examples
        ========

        >>> from sympy.abc import x, y, z
        >>> from sympy import sin
        >>> from sympy.plotting.experimental_lambdify import Lambdifier
        >>> str2tree = Lambdifier([x], x).str2tree
        >>> tree2str = Lambdifier([x], x).tree2str

        >>> tree2str(str2tree(str(x+y*sin(z)+1)))
        'x + y*sin(z) + 1'
        """
        if isinstance(tree, str):
            return tree
        else:
            return ''.join(map(cls.tree2str, tree))

    def tree2str_translate(self, tree):
        """Converts a tree to string with translations.

        Explanation
        ===========

        Function names are translated by translate_func.
        Other strings are translated by translate_str.
        """
        if isinstance(tree, str):
            return self.translate_str(tree)
        elif isinstance(tree, tuple) and len(tree) == 2:
            return self.translate_func(tree[0][:-1], tree[1])
        else:
            return ''.join([self.tree2str_translate(t) for t in tree])

    def translate_str(self, estr):
        """Translate substrings of estr using in order the dictionaries in
        dict_tuple_str."""
        for pattern, repl in self.dict_str.items():
                estr = re.sub(pattern, repl, estr)
        return estr

    def translate_func(self, func_name, argtree):
        """Translate function names and the tree of arguments.

        Explanation
        ===========

        If the function name is not in the dictionaries of dict_tuple_fun then the
        function is surrounded by a float((...).evalf()).

        The use of float is necessary as np.<function>(sympy.Float(..)) raises an
        error."""
        if func_name in self.dict_fun:
            new_name = self.dict_fun[func_name]
            argstr = self.tree2str_translate(argtree)
            return new_name + '(' + argstr
        elif func_name in ['Eq', 'Ne']:
            op = {'Eq': '==', 'Ne': '!='}
            return "(lambda x, y: x {} y)({}".format(op[func_name], self.tree2str_translate(argtree))
        else:
            template = '(%s(%s)).evalf(' if self.use_evalf else '%s(%s'
            if self.float_wrap_evalf:
                template = 'float(%s)' % template
            elif self.complex_wrap_evalf:
                template = 'complex(%s)' % template

            # Wrapping should only happen on the outermost expression, which
            # is the only thing we know will be a number.
            float_wrap_evalf = self.float_wrap_evalf
            complex_wrap_evalf = self.complex_wrap_evalf
            self.float_wrap_evalf = False
            self.complex_wrap_evalf = False
            ret =  template % (func_name, self.tree2str_translate(argtree))
            self.float_wrap_evalf = float_wrap_evalf
            self.complex_wrap_evalf = complex_wrap_evalf
            return ret

    ##############################################################################
    # The namespace constructors
    ##############################################################################

    @classmethod
    def sympy_expression_namespace(cls, expr):
        """Traverses the (func, args) tree of an expression and creates a SymPy
        namespace. All other modules are imported only as a module name. That way
        the namespace is not polluted and rests quite small. It probably causes much
        more variable lookups and so it takes more time, but there are no tests on
        that for the moment."""
        if expr is None:
            return {}
        else:
            funcname = str(expr.func)
            # XXX Workaround
            # Here we add an ugly workaround because str(func(x))
            # is not always the same as str(func). Eg
            # >>> str(Integral(x))
            # "Integral(x)"
            # >>> str(Integral)
            # "<class 'sympy.integrals.integrals.Integral'>"
            # >>> str(sqrt(x))
            # "sqrt(x)"
            # >>> str(sqrt)
            # "<function sqrt at 0x3d92de8>"
            # >>> str(sin(x))
            # "sin(x)"
            # >>> str(sin)
            # "sin"
            # Either one of those can be used but not all at the same time.
            # The code considers the sin example as the right one.
            regexlist = [
                r'<class \'sympy[\w.]*?.([\w]*)\'>$',
                # the example Integral
                r'<function ([\w]*) at 0x[\w]*>$',    # the example sqrt
            ]
            for r in regexlist:
                m = re.match(r, funcname)
                if m is not None:
                    funcname = m.groups()[0]
            # End of the workaround
            # XXX debug: print funcname
            args_dict = {}
            for a in expr.args:
                if (isinstance(a, Symbol) or
                    isinstance(a, NumberSymbol) or
                        a in [I, zoo, oo]):
                    continue
                else:
                    args_dict.update(cls.sympy_expression_namespace(a))
            args_dict.update({funcname: expr.func})
            return args_dict

    @staticmethod
    def sympy_atoms_namespace(expr):
        """For no real reason this function is separated from
        sympy_expression_namespace. It can be moved to it."""
        atoms = expr.atoms(Symbol, NumberSymbol, I, zoo, oo)
        d = {}
        for a in atoms:
            # XXX debug: print 'atom:' + str(a)
            d[str(a)] = a
        return d



from sympy.core.function import Function
from sympy.core.mul import Mul
from sympy.core.numbers import (E, I, Rational, oo, pi)
from sympy.core.singleton import S
from sympy.core.symbol import (Dummy, Symbol, symbols)
from sympy.functions.elementary.exponential import (exp, log)
from sympy.functions.elementary.miscellaneous import (root, sqrt)
from sympy.functions.elementary.trigonometric import sin
from sympy.functions.special.gamma_functions import gamma
from sympy.functions.special.hyper import hyper
from sympy.matrices.expressions.matexpr import MatrixSymbol
from sympy.simplify.powsimp import (powdenest, powsimp)
from sympy.simplify.simplify import (signsimp, simplify)
from sympy.core.symbol import Str

from sympy.abc import x, y, z, a, b


def test_powsimp():
    x, y, z, n = symbols('x,y,z,n')
    f = Function('f')
    assert powsimp( 4**x * 2**(-x) * 2**(-x) ) == 1
    assert powsimp( (-4)**x * (-2)**(-x) * 2**(-x) ) == 1

    assert powsimp(
        f(4**x * 2**(-x) * 2**(-x)) ) == f(4**x * 2**(-x) * 2**(-x))
    assert powsimp( f(4**x * 2**(-x) * 2**(-x)), deep=True ) == f(1)
    assert exp(x)*exp(y) == exp(x)*exp(y)
    assert powsimp(exp(x)*exp(y)) == exp(x + y)
    assert powsimp(exp(x)*exp(y)*2**x*2**y) == (2*E)**(x + y)
    assert powsimp(exp(x)*exp(y)*2**x*2**y, combine='exp') == \
        exp(x + y)*2**(x + y)
    assert powsimp(exp(x)*exp(y)*exp(2)*sin(x) + sin(y) + 2**x*2**y) == \
        exp(2 + x + y)*sin(x) + sin(y) + 2**(x + y)
    assert powsimp(sin(exp(x)*exp(y))) == sin(exp(x)*exp(y))
    assert powsimp(sin(exp(x)*exp(y)), deep=True) == sin(exp(x + y))
    assert powsimp(x**2*x**y) == x**(2 + y)
    # This should remain factored, because 'exp' with deep=True is supposed
    # to act like old automatic exponent combining.
    assert powsimp((1 + E*exp(E))*exp(-E), combine='exp', deep=True) == \
        (1 + exp(1 + E))*exp(-E)
    assert powsimp((1 + E*exp(E))*exp(-E), deep=True) == \
        (1 + exp(1 + E))*exp(-E)
    assert powsimp((1 + E*exp(E))*exp(-E)) == (1 + exp(1 + E))*exp(-E)
    assert powsimp((1 + E*exp(E))*exp(-E), combine='exp') == \
        (1 + exp(1 + E))*exp(-E)
    assert powsimp((1 + E*exp(E))*exp(-E), combine='base') == \
        (1 + E*exp(E))*exp(-E)
    x, y = symbols('x,y', nonnegative=True)
    n = Symbol('n', real=True)
    assert powsimp(y**n * (y/x)**(-n)) == x**n
    assert powsimp(x**(x**(x*y)*y**(x*y))*y**(x**(x*y)*y**(x*y)), deep=True) \
        == (x*y)**(x*y)**(x*y)
    assert powsimp(2**(2**(2*x)*x), deep=False) == 2**(2**(2*x)*x)
    assert powsimp(2**(2**(2*x)*x), deep=True) == 2**(x*4**x)
    assert powsimp(
        exp(-x + exp(-x)*exp(-x*log(x))), deep=False, combine='exp') == \
        exp(-x + exp(-x)*exp(-x*log(x)))
    assert powsimp(
        exp(-x + exp(-x)*exp(-x*log(x))), deep=False, combine='exp') == \
        exp(-x + exp(-x)*exp(-x*log(x)))
    assert powsimp((x + y)/(3*z), deep=False, combine='exp') == (x + y)/(3*z)
    assert powsimp((x/3 + y/3)/z, deep=True, combine='exp') == (x/3 + y/3)/z
    assert powsimp(exp(x)/(1 + exp(x)*exp(y)), deep=True) == \
        exp(x)/(1 + exp(x + y))
    assert powsimp(x*y**(z**x*z**y), deep=True) == x*y**(z**(x + y))
    assert powsimp((z**x*z**y)**x, deep=True) == (z**(x + y))**x
    assert powsimp(x*(z**x*z**y)**x, deep=True) == x*(z**(x + y))**x
    p = symbols('p', positive=True)
    assert powsimp((1/x)**log(2)/x) == (1/x)**(1 + log(2))
    assert powsimp((1/p)**log(2)/p) == p**(-1 - log(2))

    # coefficient of exponent can only be simplified for positive bases
    assert powsimp(2**(2*x)) == 4**x
    assert powsimp((-1)**(2*x)) == (-1)**(2*x)
    i = symbols('i', integer=True)
    assert powsimp((-1)**(2*i)) == 1
    assert powsimp((-1)**(-x)) != (-1)**x  # could be 1/((-1)**x), but is not
    # force=True overrides assumptions
    assert powsimp((-1)**(2*x), force=True) == 1

    # rational exponents allow combining of negative terms
    w, n, m = symbols('w n m', negative=True)
    e = i/a  # not a rational exponent if `a` is unknown
    ex = w**e*n**e*m**e
    assert powsimp(ex) == m**(i/a)*n**(i/a)*w**(i/a)
    e = i/3
    ex = w**e*n**e*m**e
    assert powsimp(ex) == (-1)**i*(-m*n*w)**(i/3)
    e = (3 + i)/i
    ex = w**e*n**e*m**e
    assert powsimp(ex) == (-1)**(3*e)*(-m*n*w)**e

    eq = x**(a*Rational(2, 3))
    # eq != (x**a)**(2/3) (try x = -1 and a = 3 to see)
    assert powsimp(eq).exp == eq.exp == a*Rational(2, 3)
    # powdenest goes the other direction
    assert powsimp(2**(2*x)) == 4**x

    assert powsimp(exp(p/2)) == exp(p/2)

    # issue 6368
    eq = Mul(*[sqrt(Dummy(imaginary=True)) for i in range(3)])
    assert powsimp(eq) == eq and eq.is_Mul

    assert all(powsimp(e) == e for e in (sqrt(x**a), sqrt(x**2)))

    # issue 8836
    assert str( powsimp(exp(I*pi/3)*root(-1,3)) ) == '(-1)**(2/3)'

    # issue 9183
    assert powsimp(-0.1**x) == -0.1**x

    # issue 10095
    assert powsimp((1/(2*E))**oo) == (exp(-1)/2)**oo

    # PR 13131
    eq = sin(2*x)**2*sin(2.0*x)**2
    assert powsimp(eq) == eq

    # issue 14615
    assert powsimp(x**2*y**3*(x*y**2)**Rational(3, 2)
        ) == x*y*(x*y**2)**Rational(5, 2)


def test_powsimp_negated_base():
    assert powsimp((-x + y)/sqrt(x - y)) == -sqrt(x - y)
    assert powsimp((-x + y)*(-z + y)/sqrt(x - y)/sqrt(z - y)) == sqrt(x - y)*sqrt(z - y)
    p = symbols('p', positive=True)
    reps = {p: 2, a: S.Half}
    assert powsimp((-p)**a/p**a).subs(reps) == ((-1)**a).subs(reps)
    assert powsimp((-p)**a*p**a).subs(reps) == ((-p**2)**a).subs(reps)
    n = symbols('n', negative=True)
    reps = {p: -2, a: S.Half}
    assert powsimp((-n)**a/n**a).subs(reps) == (-1)**(-a).subs(a, S.Half)
    assert powsimp((-n)**a*n**a).subs(reps) == ((-n**2)**a).subs(reps)
    # if x is 0 then the lhs is 0**a*oo**a which is not (-1)**a
    eq = (-x)**a/x**a
    assert powsimp(eq) == eq


def test_powsimp_nc():
    x, y, z = symbols('x,y,z')
    A, B, C = symbols('A B C', commutative=False)

    assert powsimp(A**x*A**y, combine='all') == A**(x + y)
    assert powsimp(A**x*A**y, combine='base') == A**x*A**y
    assert powsimp(A**x*A**y, combine='exp') == A**(x + y)

    assert powsimp(A**x*B**x, combine='all') == A**x*B**x
    assert powsimp(A**x*B**x, combine='base') == A**x*B**x
    assert powsimp(A**x*B**x, combine='exp') == A**x*B**x

    assert powsimp(B**x*A**x, combine='all') == B**x*A**x
    assert powsimp(B**x*A**x, combine='base') == B**x*A**x
    assert powsimp(B**x*A**x, combine='exp') == B**x*A**x

    assert powsimp(A**x*A**y*A**z, combine='all') == A**(x + y + z)
    assert powsimp(A**x*A**y*A**z, combine='base') == A**x*A**y*A**z
    assert powsimp(A**x*A**y*A**z, combine='exp') == A**(x + y + z)

    assert powsimp(A**x*B**x*C**x, combine='all') == A**x*B**x*C**x
    assert powsimp(A**x*B**x*C**x, combine='base') == A**x*B**x*C**x
    assert powsimp(A**x*B**x*C**x, combine='exp') == A**x*B**x*C**x

    assert powsimp(B**x*A**x*C**x, combine='all') == B**x*A**x*C**x
    assert powsimp(B**x*A**x*C**x, combine='base') == B**x*A**x*C**x
    assert powsimp(B**x*A**x*C**x, combine='exp') == B**x*A**x*C**x


def test_issue_6440():
    assert powsimp(16*2**a*8**b) == 2**(a + 3*b + 4)


def test_powdenest():
    x, y = symbols('x,y')
    p, q = symbols('p q', positive=True)
    i, j = symbols('i,j', integer=True)

    assert powdenest(x) == x
    assert powdenest(x + 2*(x**(a*Rational(2, 3)))**(3*x)) == (x + 2*(x**(a*Rational(2, 3)))**(3*x))
    assert powdenest((exp(a*Rational(2, 3)))**(3*x))  # -X-> (exp(a/3))**(6*x)
    assert powdenest((x**(a*Rational(2, 3)))**(3*x)) == ((x**(a*Rational(2, 3)))**(3*x))
    assert powdenest(exp(3*x*log(2))) == 2**(3*x)
    assert powdenest(sqrt(p**2)) == p
    eq = p**(2*i)*q**(4*i)
    assert powdenest(eq) == (p*q**2)**(2*i)
    # -X-> (x**x)**i*(x**x)**j == x**(x*(i + j))
    assert powdenest((x**x)**(i + j))
    assert powdenest(exp(3*y*log(x))) == x**(3*y)
    assert powdenest(exp(y*(log(a) + log(b)))) == (a*b)**y
    assert powdenest(exp(3*(log(a) + log(b)))) == a**3*b**3
    assert powdenest(((x**(2*i))**(3*y))**x) == ((x**(2*i))**(3*y))**x
    assert powdenest(((x**(2*i))**(3*y))**x, force=True) == x**(6*i*x*y)
    assert powdenest(((x**(a*Rational(2, 3)))**(3*y/i))**x) == \
        (((x**(a*Rational(2, 3)))**(3*y/i))**x)
    assert powdenest((x**(2*i)*y**(4*i))**z, force=True) == (x*y**2)**(2*i*z)
    assert powdenest((p**(2*i)*q**(4*i))**j) == (p*q**2)**(2*i*j)
    e = ((p**(2*a))**(3*y))**x
    assert powdenest(e) == e
    e = ((x**2*y**4)**a)**(x*y)
    assert powdenest(e) == e
    e = (((x**2*y**4)**a)**(x*y))**3
    assert powdenest(e) == ((x**2*y**4)**a)**(3*x*y)
    assert powdenest((((x**2*y**4)**a)**(x*y)), force=True) == \
        (x*y**2)**(2*a*x*y)
    assert powdenest((((x**2*y**4)**a)**(x*y))**3, force=True) == \
        (x*y**2)**(6*a*x*y)
    assert powdenest((x**2*y**6)**i) != (x*y**3)**(2*i)
    x, y = symbols('x,y', positive=True)
    assert powdenest((x**2*y**6)**i) == (x*y**3)**(2*i)

    assert powdenest((x**(i*Rational(2, 3))*y**(i/2))**(2*i)) == (x**Rational(4, 3)*y)**(i**2)
    assert powdenest(sqrt(x**(2*i)*y**(6*i))) == (x*y**3)**i

    assert powdenest(4**x) == 2**(2*x)
    assert powdenest((4**x)**y) == 2**(2*x*y)
    assert powdenest(4**x*y) == 2**(2*x)*y


def test_powdenest_polar():
    x, y, z = symbols('x y z', polar=True)
    a, b, c = symbols('a b c')
    assert powdenest((x*y*z)**a) == x**a*y**a*z**a
    assert powdenest((x**a*y**b)**c) == x**(a*c)*y**(b*c)
    assert powdenest(((x**a)**b*y**c)**c) == x**(a*b*c)*y**(c**2)


def test_issue_5805():
    arg = ((gamma(x)*hyper((), (), x))*pi)**2
    assert powdenest(arg) == (pi*gamma(x)*hyper((), (), x))**2
    assert arg.is_positive is None


def test_issue_9324_powsimp_on_matrix_symbol():
    M = MatrixSymbol('M', 10, 10)
    expr = powsimp(M, deep=True)
    assert expr == M
    assert expr.args[0] == Str('M')


def test_issue_6367():
    z = -5*sqrt(2)/(2*sqrt(2*sqrt(29) + 29)) + sqrt(-sqrt(29)/29 + S.Half)
    assert Mul(*[powsimp(a) for a in Mul.make_args(z.normal())]) == 0
    assert powsimp(z.normal()) == 0
    assert simplify(z) == 0
    assert powsimp(sqrt(2 + sqrt(3))*sqrt(2 - sqrt(3)) + 1) == 2
    assert powsimp(z) != 0


def test_powsimp_polar():
    from sympy.functions.elementary.complexes import polar_lift
    from sympy.functions.elementary.exponential import exp_polar
    x, y, z = symbols('x y z')
    p, q, r = symbols('p q r', polar=True)

    assert (polar_lift(-1))**(2*x) == exp_polar(2*pi*I*x)
    assert powsimp(p**x * q**x) == (p*q)**x
    assert p**x * (1/p)**x == 1
    assert (1/p)**x == p**(-x)

    assert exp_polar(x)*exp_polar(y) == exp_polar(x)*exp_polar(y)
    assert powsimp(exp_polar(x)*exp_polar(y)) == exp_polar(x + y)
    assert powsimp(exp_polar(x)*exp_polar(y)*p**x*p**y) == \
        (p*exp_polar(1))**(x + y)
    assert powsimp(exp_polar(x)*exp_polar(y)*p**x*p**y, combine='exp') == \
        exp_polar(x + y)*p**(x + y)
    assert powsimp(
        exp_polar(x)*exp_polar(y)*exp_polar(2)*sin(x) + sin(y) + p**x*p**y) \
        == p**(x + y) + sin(x)*exp_polar(2 + x + y) + sin(y)
    assert powsimp(sin(exp_polar(x)*exp_polar(y))) == \
        sin(exp_polar(x)*exp_polar(y))
    assert powsimp(sin(exp_polar(x)*exp_polar(y)), deep=True) == \
        sin(exp_polar(x + y))


def test_issue_5728():
    b = x*sqrt(y)
    a = sqrt(b)
    c = sqrt(sqrt(x)*y)
    assert powsimp(a*b) == sqrt(b)**3
    assert powsimp(a*b**2*sqrt(y)) == sqrt(y)*a**5
    assert powsimp(a*x**2*c**3*y) == c**3*a**5
    assert powsimp(a*x*c**3*y**2) == c**7*a
    assert powsimp(x*c**3*y**2) == c**7
    assert powsimp(x*c**3*y) == x*y*c**3
    assert powsimp(sqrt(x)*c**3*y) == c**5
    assert powsimp(sqrt(x)*a**3*sqrt(y)) == sqrt(x)*sqrt(y)*a**3
    assert powsimp(Mul(sqrt(x)*c**3*sqrt(y), y, evaluate=False)) == \
        sqrt(x)*sqrt(y)**3*c**3
    assert powsimp(a**2*a*x**2*y) == a**7

    # symbolic powers work, too
    b = x**y*y
    a = b*sqrt(b)
    assert a.is_Mul is True
    assert powsimp(a) == sqrt(b)**3

    # as does exp
    a = x*exp(y*Rational(2, 3))
    assert powsimp(a*sqrt(a)) == sqrt(a)**3
    assert powsimp(a**2*sqrt(a)) == sqrt(a)**5
    assert powsimp(a**2*sqrt(sqrt(a))) == sqrt(sqrt(a))**9


def test_issue_from_PR1599():
    n1, n2, n3, n4 = symbols('n1 n2 n3 n4', negative=True)
    assert (powsimp(sqrt(n1)*sqrt(n2)*sqrt(n3)) ==
        -I*sqrt(-n1)*sqrt(-n2)*sqrt(-n3))
    assert (powsimp(root(n1, 3)*root(n2, 3)*root(n3, 3)*root(n4, 3)) ==
        -(-1)**Rational(1, 3)*
        (-n1)**Rational(1, 3)*(-n2)**Rational(1, 3)*(-n3)**Rational(1, 3)*(-n4)**Rational(1, 3))


def test_issue_10195():
    a = Symbol('a', integer=True)
    l = Symbol('l', even=True, nonzero=True)
    n = Symbol('n', odd=True)
    e_x = (-1)**(n/2 - S.Half) - (-1)**(n*Rational(3, 2) - S.Half)
    assert powsimp((-1)**(l/2)) == I**l
    assert powsimp((-1)**(n/2)) == I**n
    assert powsimp((-1)**(n*Rational(3, 2))) == -I**n
    assert powsimp(e_x) == (-1)**(n/2 - S.Half) + (-1)**(n*Rational(3, 2) +
            S.Half)
    assert powsimp((-1)**(a*Rational(3, 2))) == (-I)**a

def test_issue_15709():
    assert powsimp(3**x*Rational(2, 3)) == 2*3**(x-1)
    assert powsimp(2*3**x/3) == 2*3**(x-1)


def test_issue_11981():
    x, y = symbols('x y', commutative=False)
    assert powsimp((x*y)**2 * (y*x)**2) == (x*y)**2 * (y*x)**2


def test_issue_17524():
    a = symbols("a", real=True)
    e = (-1 - a**2)*sqrt(1 + a**2)
    assert signsimp(powsimp(e)) == signsimp(e) == -(a**2 + 1)**(S(3)/2)


def test_issue_19627():
    # if you use force the user must verify
    assert powdenest(sqrt(sin(x)**2), force=True) == sin(x)
    assert powdenest((x**(S.Half/y))**(2*y), force=True) == x
    from sympy.core.function import expand_power_base
    e = 1 - a
    expr = (exp(z/e)*x**(b/e)*y**((1 - b)/e))**e
    assert powdenest(expand_power_base(expr, force=True), force=True
        ) == x**b*y**(1 - b)*exp(z)


def test_issue_22546():
    p1, p2 = symbols('p1, p2', positive=True)
    ref = powsimp(p1**z/p2**z)
    e = z + 1
    ans = ref.subs(z, e)
    assert ans.is_Pow
    assert powsimp(p1**e/p2**e) == ans
    i = symbols('i', integer=True)
    ref = powsimp(x**i/y**i)
    e = i + 1
    ans = ref.subs(i, e)
    assert ans.is_Pow
    assert powsimp(x**e/y**e) == ans



from sympy.core.expr import Expr
from sympy.core.mul import Mul
from sympy.core.numbers import (I, Integer)
from sympy.core.symbol import symbols
from sympy.functions.elementary.complexes import conjugate
from sympy.matrices.dense import Matrix

from sympy.physics.quantum.dagger import adjoint, Dagger
from sympy.external import import_module
from sympy.testing.pytest import skip
from sympy.physics.quantum.operator import Operator, IdentityOperator


def test_scalars():
    x = symbols('x', complex=True)
    assert Dagger(x) == conjugate(x)
    assert Dagger(I*x) == -I*conjugate(x)

    i = symbols('i', real=True)
    assert Dagger(i) == i

    p = symbols('p')
    assert isinstance(Dagger(p), adjoint)

    i = Integer(3)
    assert Dagger(i) == i

    A = symbols('A', commutative=False)
    assert Dagger(A).is_commutative is False


def test_matrix():
    x = symbols('x')
    m = Matrix([[I, x*I], [2, 4]])
    assert Dagger(m) == m.H


def test_dagger_mul():
    O = Operator('O')
    I = IdentityOperator()
    assert Dagger(O)*O == Dagger(O)*O
    assert Dagger(O)*O*I == Mul(Dagger(O), O)*I
    assert Dagger(O)*Dagger(O) == Dagger(O)**2
    assert Dagger(O)*Dagger(I) == Dagger(O)


class Foo(Expr):

    def _eval_adjoint(self):
        return I


def test_eval_adjoint():
    f = Foo()
    d = Dagger(f)
    assert d == I

np = import_module('numpy')


def test_numpy_dagger():
    if not np:
        skip("numpy not installed.")

    a = np.array([[1.0, 2.0j], [-1.0j, 2.0]])
    adag = a.copy().transpose().conjugate()
    assert (Dagger(a) == adag).all()


scipy = import_module('scipy', import_kwargs={'fromlist': ['sparse']})


def test_scipy_sparse_dagger():
    if not np:
        skip("numpy not installed.")
    if not scipy:
        skip("scipy not installed.")
    else:
        sparse = scipy.sparse

    a = sparse.csr_matrix([[1.0 + 0.0j, 2.0j], [-1.0j, 2.0 + 0.0j]])
    adag = a.copy().transpose().conjugate()
    assert np.linalg.norm((Dagger(a) - adag).todense()) == 0.0



from sympy.testing.pytest import raises
from sympy.utilities.exceptions import sympy_deprecation_warning

# Only test exceptions here because the other cases are tested in the
# warns_deprecated_sympy tests
def test_sympy_deprecation_warning():
    raises(TypeError, lambda: sympy_deprecation_warning('test',
                                                        deprecated_since_version=1.10,
                                                        active_deprecations_target='active-deprecations'))

    raises(ValueError, lambda: sympy_deprecation_warning('test',
                                                            deprecated_since_version="1.10", active_deprecations_target='(active-deprecations)='))



"""
Do NOT manually edit this file.
Instead, run ./bin/ask_update.py.
"""

defined_facts = [
    'algebraic',
    'antihermitian',
    'commutative',
    'complex',
    'composite',
    'even',
    'extended_negative',
    'extended_nonnegative',
    'extended_nonpositive',
    'extended_nonzero',
    'extended_positive',
    'extended_real',
    'finite',
    'hermitian',
    'imaginary',
    'infinite',
    'integer',
    'irrational',
    'negative',
    'noninteger',
    'nonnegative',
    'nonpositive',
    'nonzero',
    'odd',
    'positive',
    'prime',
    'rational',
    'real',
    'transcendental',
    'zero',
] # defined_facts


full_implications = dict( [
    # Implications of algebraic = True:
    (('algebraic', True), set( (
        ('commutative', True),
        ('complex', True),
        ('finite', True),
        ('infinite', False),
        ('transcendental', False),
       ) ),
     ),
    # Implications of algebraic = False:
    (('algebraic', False), set( (
        ('composite', False),
        ('even', False),
        ('integer', False),
        ('odd', False),
        ('prime', False),
        ('rational', False),
        ('zero', False),
       ) ),
     ),
    # Implications of antihermitian = True:
    (('antihermitian', True), set( (
       ) ),
     ),
    # Implications of antihermitian = False:
    (('antihermitian', False), set( (
        ('imaginary', False),
       ) ),
     ),
    # Implications of commutative = True:
    (('commutative', True), set( (
       ) ),
     ),
    # Implications of commutative = False:
    (('commutative', False), set( (
        ('algebraic', False),
        ('complex', False),
        ('composite', False),
        ('even', False),
        ('extended_negative', False),
        ('extended_nonnegative', False),
        ('extended_nonpositive', False),
        ('extended_nonzero', False),
        ('extended_positive', False),
        ('extended_real', False),
        ('imaginary', False),
        ('integer', False),
        ('irrational', False),
        ('negative', False),
        ('noninteger', False),
        ('nonnegative', False),
        ('nonpositive', False),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', False),
        ('real', False),
        ('transcendental', False),
        ('zero', False),
       ) ),
     ),
    # Implications of complex = True:
    (('complex', True), set( (
        ('commutative', True),
        ('finite', True),
        ('infinite', False),
       ) ),
     ),
    # Implications of complex = False:
    (('complex', False), set( (
        ('algebraic', False),
        ('composite', False),
        ('even', False),
        ('imaginary', False),
        ('integer', False),
        ('irrational', False),
        ('negative', False),
        ('nonnegative', False),
        ('nonpositive', False),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', False),
        ('real', False),
        ('transcendental', False),
        ('zero', False),
       ) ),
     ),
    # Implications of composite = True:
    (('composite', True), set( (
        ('algebraic', True),
        ('commutative', True),
        ('complex', True),
        ('extended_negative', False),
        ('extended_nonnegative', True),
        ('extended_nonpositive', False),
        ('extended_nonzero', True),
        ('extended_positive', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('integer', True),
        ('irrational', False),
        ('negative', False),
        ('noninteger', False),
        ('nonnegative', True),
        ('nonpositive', False),
        ('nonzero', True),
        ('positive', True),
        ('prime', False),
        ('rational', True),
        ('real', True),
        ('transcendental', False),
        ('zero', False),
       ) ),
     ),
    # Implications of composite = False:
    (('composite', False), set( (
       ) ),
     ),
    # Implications of even = True:
    (('even', True), set( (
        ('algebraic', True),
        ('commutative', True),
        ('complex', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('integer', True),
        ('irrational', False),
        ('noninteger', False),
        ('odd', False),
        ('rational', True),
        ('real', True),
        ('transcendental', False),
       ) ),
     ),
    # Implications of even = False:
    (('even', False), set( (
        ('zero', False),
       ) ),
     ),
    # Implications of extended_negative = True:
    (('extended_negative', True), set( (
        ('commutative', True),
        ('composite', False),
        ('extended_nonnegative', False),
        ('extended_nonpositive', True),
        ('extended_nonzero', True),
        ('extended_positive', False),
        ('extended_real', True),
        ('imaginary', False),
        ('nonnegative', False),
        ('positive', False),
        ('prime', False),
        ('zero', False),
       ) ),
     ),
    # Implications of extended_negative = False:
    (('extended_negative', False), set( (
        ('negative', False),
       ) ),
     ),
    # Implications of extended_nonnegative = True:
    (('extended_nonnegative', True), set( (
        ('commutative', True),
        ('extended_negative', False),
        ('extended_real', True),
        ('imaginary', False),
        ('negative', False),
       ) ),
     ),
    # Implications of extended_nonnegative = False:
    (('extended_nonnegative', False), set( (
        ('composite', False),
        ('extended_positive', False),
        ('nonnegative', False),
        ('positive', False),
        ('prime', False),
        ('zero', False),
       ) ),
     ),
    # Implications of extended_nonpositive = True:
    (('extended_nonpositive', True), set( (
        ('commutative', True),
        ('composite', False),
        ('extended_positive', False),
        ('extended_real', True),
        ('imaginary', False),
        ('positive', False),
        ('prime', False),
       ) ),
     ),
    # Implications of extended_nonpositive = False:
    (('extended_nonpositive', False), set( (
        ('extended_negative', False),
        ('negative', False),
        ('nonpositive', False),
        ('zero', False),
       ) ),
     ),
    # Implications of extended_nonzero = True:
    (('extended_nonzero', True), set( (
        ('commutative', True),
        ('extended_real', True),
        ('imaginary', False),
        ('zero', False),
       ) ),
     ),
    # Implications of extended_nonzero = False:
    (('extended_nonzero', False), set( (
        ('composite', False),
        ('extended_negative', False),
        ('extended_positive', False),
        ('negative', False),
        ('nonzero', False),
        ('positive', False),
        ('prime', False),
       ) ),
     ),
    # Implications of extended_positive = True:
    (('extended_positive', True), set( (
        ('commutative', True),
        ('extended_negative', False),
        ('extended_nonnegative', True),
        ('extended_nonpositive', False),
        ('extended_nonzero', True),
        ('extended_real', True),
        ('imaginary', False),
        ('negative', False),
        ('nonpositive', False),
        ('zero', False),
       ) ),
     ),
    # Implications of extended_positive = False:
    (('extended_positive', False), set( (
        ('composite', False),
        ('positive', False),
        ('prime', False),
       ) ),
     ),
    # Implications of extended_real = True:
    (('extended_real', True), set( (
        ('commutative', True),
        ('imaginary', False),
       ) ),
     ),
    # Implications of extended_real = False:
    (('extended_real', False), set( (
        ('composite', False),
        ('even', False),
        ('extended_negative', False),
        ('extended_nonnegative', False),
        ('extended_nonpositive', False),
        ('extended_nonzero', False),
        ('extended_positive', False),
        ('integer', False),
        ('irrational', False),
        ('negative', False),
        ('noninteger', False),
        ('nonnegative', False),
        ('nonpositive', False),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', False),
        ('real', False),
        ('zero', False),
       ) ),
     ),
    # Implications of finite = True:
    (('finite', True), set( (
        ('infinite', False),
       ) ),
     ),
    # Implications of finite = False:
    (('finite', False), set( (
        ('algebraic', False),
        ('complex', False),
        ('composite', False),
        ('even', False),
        ('imaginary', False),
        ('infinite', True),
        ('integer', False),
        ('irrational', False),
        ('negative', False),
        ('nonnegative', False),
        ('nonpositive', False),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', False),
        ('real', False),
        ('transcendental', False),
        ('zero', False),
       ) ),
     ),
    # Implications of hermitian = True:
    (('hermitian', True), set( (
       ) ),
     ),
    # Implications of hermitian = False:
    (('hermitian', False), set( (
        ('composite', False),
        ('even', False),
        ('integer', False),
        ('irrational', False),
        ('negative', False),
        ('nonnegative', False),
        ('nonpositive', False),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', False),
        ('real', False),
        ('zero', False),
       ) ),
     ),
    # Implications of imaginary = True:
    (('imaginary', True), set( (
        ('antihermitian', True),
        ('commutative', True),
        ('complex', True),
        ('composite', False),
        ('even', False),
        ('extended_negative', False),
        ('extended_nonnegative', False),
        ('extended_nonpositive', False),
        ('extended_nonzero', False),
        ('extended_positive', False),
        ('extended_real', False),
        ('finite', True),
        ('infinite', False),
        ('integer', False),
        ('irrational', False),
        ('negative', False),
        ('noninteger', False),
        ('nonnegative', False),
        ('nonpositive', False),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', False),
        ('real', False),
        ('zero', False),
       ) ),
     ),
    # Implications of imaginary = False:
    (('imaginary', False), set( (
       ) ),
     ),
    # Implications of infinite = True:
    (('infinite', True), set( (
        ('algebraic', False),
        ('complex', False),
        ('composite', False),
        ('even', False),
        ('finite', False),
        ('imaginary', False),
        ('integer', False),
        ('irrational', False),
        ('negative', False),
        ('nonnegative', False),
        ('nonpositive', False),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', False),
        ('real', False),
        ('transcendental', False),
        ('zero', False),
       ) ),
     ),
    # Implications of infinite = False:
    (('infinite', False), set( (
        ('finite', True),
       ) ),
     ),
    # Implications of integer = True:
    (('integer', True), set( (
        ('algebraic', True),
        ('commutative', True),
        ('complex', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('irrational', False),
        ('noninteger', False),
        ('rational', True),
        ('real', True),
        ('transcendental', False),
       ) ),
     ),
    # Implications of integer = False:
    (('integer', False), set( (
        ('composite', False),
        ('even', False),
        ('odd', False),
        ('prime', False),
        ('zero', False),
       ) ),
     ),
    # Implications of irrational = True:
    (('irrational', True), set( (
        ('commutative', True),
        ('complex', True),
        ('composite', False),
        ('even', False),
        ('extended_nonzero', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('integer', False),
        ('noninteger', True),
        ('nonzero', True),
        ('odd', False),
        ('prime', False),
        ('rational', False),
        ('real', True),
        ('zero', False),
       ) ),
     ),
    # Implications of irrational = False:
    (('irrational', False), set( (
       ) ),
     ),
    # Implications of negative = True:
    (('negative', True), set( (
        ('commutative', True),
        ('complex', True),
        ('composite', False),
        ('extended_negative', True),
        ('extended_nonnegative', False),
        ('extended_nonpositive', True),
        ('extended_nonzero', True),
        ('extended_positive', False),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('nonnegative', False),
        ('nonpositive', True),
        ('nonzero', True),
        ('positive', False),
        ('prime', False),
        ('real', True),
        ('zero', False),
       ) ),
     ),
    # Implications of negative = False:
    (('negative', False), set( (
       ) ),
     ),
    # Implications of noninteger = True:
    (('noninteger', True), set( (
        ('commutative', True),
        ('composite', False),
        ('even', False),
        ('extended_nonzero', True),
        ('extended_real', True),
        ('imaginary', False),
        ('integer', False),
        ('odd', False),
        ('prime', False),
        ('zero', False),
       ) ),
     ),
    # Implications of noninteger = False:
    (('noninteger', False), set( (
       ) ),
     ),
    # Implications of nonnegative = True:
    (('nonnegative', True), set( (
        ('commutative', True),
        ('complex', True),
        ('extended_negative', False),
        ('extended_nonnegative', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('negative', False),
        ('real', True),
       ) ),
     ),
    # Implications of nonnegative = False:
    (('nonnegative', False), set( (
        ('composite', False),
        ('positive', False),
        ('prime', False),
        ('zero', False),
       ) ),
     ),
    # Implications of nonpositive = True:
    (('nonpositive', True), set( (
        ('commutative', True),
        ('complex', True),
        ('composite', False),
        ('extended_nonpositive', True),
        ('extended_positive', False),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('positive', False),
        ('prime', False),
        ('real', True),
       ) ),
     ),
    # Implications of nonpositive = False:
    (('nonpositive', False), set( (
        ('negative', False),
        ('zero', False),
       ) ),
     ),
    # Implications of nonzero = True:
    (('nonzero', True), set( (
        ('commutative', True),
        ('complex', True),
        ('extended_nonzero', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('real', True),
        ('zero', False),
       ) ),
     ),
    # Implications of nonzero = False:
    (('nonzero', False), set( (
        ('composite', False),
        ('negative', False),
        ('positive', False),
        ('prime', False),
       ) ),
     ),
    # Implications of odd = True:
    (('odd', True), set( (
        ('algebraic', True),
        ('commutative', True),
        ('complex', True),
        ('even', False),
        ('extended_nonzero', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('integer', True),
        ('irrational', False),
        ('noninteger', False),
        ('nonzero', True),
        ('rational', True),
        ('real', True),
        ('transcendental', False),
        ('zero', False),
       ) ),
     ),
    # Implications of odd = False:
    (('odd', False), set( (
       ) ),
     ),
    # Implications of positive = True:
    (('positive', True), set( (
        ('commutative', True),
        ('complex', True),
        ('extended_negative', False),
        ('extended_nonnegative', True),
        ('extended_nonpositive', False),
        ('extended_nonzero', True),
        ('extended_positive', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('negative', False),
        ('nonnegative', True),
        ('nonpositive', False),
        ('nonzero', True),
        ('real', True),
        ('zero', False),
       ) ),
     ),
    # Implications of positive = False:
    (('positive', False), set( (
        ('composite', False),
        ('prime', False),
       ) ),
     ),
    # Implications of prime = True:
    (('prime', True), set( (
        ('algebraic', True),
        ('commutative', True),
        ('complex', True),
        ('composite', False),
        ('extended_negative', False),
        ('extended_nonnegative', True),
        ('extended_nonpositive', False),
        ('extended_nonzero', True),
        ('extended_positive', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('integer', True),
        ('irrational', False),
        ('negative', False),
        ('noninteger', False),
        ('nonnegative', True),
        ('nonpositive', False),
        ('nonzero', True),
        ('positive', True),
        ('rational', True),
        ('real', True),
        ('transcendental', False),
        ('zero', False),
       ) ),
     ),
    # Implications of prime = False:
    (('prime', False), set( (
       ) ),
     ),
    # Implications of rational = True:
    (('rational', True), set( (
        ('algebraic', True),
        ('commutative', True),
        ('complex', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('irrational', False),
        ('real', True),
        ('transcendental', False),
       ) ),
     ),
    # Implications of rational = False:
    (('rational', False), set( (
        ('composite', False),
        ('even', False),
        ('integer', False),
        ('odd', False),
        ('prime', False),
        ('zero', False),
       ) ),
     ),
    # Implications of real = True:
    (('real', True), set( (
        ('commutative', True),
        ('complex', True),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
       ) ),
     ),
    # Implications of real = False:
    (('real', False), set( (
        ('composite', False),
        ('even', False),
        ('integer', False),
        ('irrational', False),
        ('negative', False),
        ('nonnegative', False),
        ('nonpositive', False),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', False),
        ('zero', False),
       ) ),
     ),
    # Implications of transcendental = True:
    (('transcendental', True), set( (
        ('algebraic', False),
        ('commutative', True),
        ('complex', True),
        ('composite', False),
        ('even', False),
        ('finite', True),
        ('infinite', False),
        ('integer', False),
        ('odd', False),
        ('prime', False),
        ('rational', False),
        ('zero', False),
       ) ),
     ),
    # Implications of transcendental = False:
    (('transcendental', False), set( (
       ) ),
     ),
    # Implications of zero = True:
    (('zero', True), set( (
        ('algebraic', True),
        ('commutative', True),
        ('complex', True),
        ('composite', False),
        ('even', True),
        ('extended_negative', False),
        ('extended_nonnegative', True),
        ('extended_nonpositive', True),
        ('extended_nonzero', False),
        ('extended_positive', False),
        ('extended_real', True),
        ('finite', True),
        ('hermitian', True),
        ('imaginary', False),
        ('infinite', False),
        ('integer', True),
        ('irrational', False),
        ('negative', False),
        ('noninteger', False),
        ('nonnegative', True),
        ('nonpositive', True),
        ('nonzero', False),
        ('odd', False),
        ('positive', False),
        ('prime', False),
        ('rational', True),
        ('real', True),
        ('transcendental', False),
       ) ),
     ),
    # Implications of zero = False:
    (('zero', False), set( (
       ) ),
     ),
 ] ) # full_implications


prereq = {

    # facts that could determine the value of algebraic
    'algebraic': {
        'commutative',
        'complex',
        'composite',
        'even',
        'finite',
        'infinite',
        'integer',
        'odd',
        'prime',
        'rational',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of antihermitian
    'antihermitian': {
        'imaginary',
    },

    # facts that could determine the value of commutative
    'commutative': {
        'algebraic',
        'complex',
        'composite',
        'even',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'extended_real',
        'imaginary',
        'integer',
        'irrational',
        'negative',
        'noninteger',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of complex
    'complex': {
        'algebraic',
        'commutative',
        'composite',
        'even',
        'finite',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'negative',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of composite
    'composite': {
        'algebraic',
        'commutative',
        'complex',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'negative',
        'noninteger',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'positive',
        'prime',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of even
    'even': {
        'algebraic',
        'commutative',
        'complex',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'noninteger',
        'odd',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of extended_negative
    'extended_negative': {
        'commutative',
        'composite',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'extended_real',
        'imaginary',
        'negative',
        'nonnegative',
        'positive',
        'prime',
        'zero',
    },

    # facts that could determine the value of extended_nonnegative
    'extended_nonnegative': {
        'commutative',
        'composite',
        'extended_negative',
        'extended_positive',
        'extended_real',
        'imaginary',
        'negative',
        'nonnegative',
        'positive',
        'prime',
        'zero',
    },

    # facts that could determine the value of extended_nonpositive
    'extended_nonpositive': {
        'commutative',
        'composite',
        'extended_negative',
        'extended_positive',
        'extended_real',
        'imaginary',
        'negative',
        'nonpositive',
        'positive',
        'prime',
        'zero',
    },

    # facts that could determine the value of extended_nonzero
    'extended_nonzero': {
        'commutative',
        'composite',
        'extended_negative',
        'extended_positive',
        'extended_real',
        'imaginary',
        'irrational',
        'negative',
        'noninteger',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'zero',
    },

    # facts that could determine the value of extended_positive
    'extended_positive': {
        'commutative',
        'composite',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_real',
        'imaginary',
        'negative',
        'nonpositive',
        'positive',
        'prime',
        'zero',
    },

    # facts that could determine the value of extended_real
    'extended_real': {
        'commutative',
        'composite',
        'even',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'imaginary',
        'integer',
        'irrational',
        'negative',
        'noninteger',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'real',
        'zero',
    },

    # facts that could determine the value of finite
    'finite': {
        'algebraic',
        'complex',
        'composite',
        'even',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'negative',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of hermitian
    'hermitian': {
        'composite',
        'even',
        'integer',
        'irrational',
        'negative',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'real',
        'zero',
    },

    # facts that could determine the value of imaginary
    'imaginary': {
        'antihermitian',
        'commutative',
        'complex',
        'composite',
        'even',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'extended_real',
        'finite',
        'infinite',
        'integer',
        'irrational',
        'negative',
        'noninteger',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'real',
        'zero',
    },

    # facts that could determine the value of infinite
    'infinite': {
        'algebraic',
        'complex',
        'composite',
        'even',
        'finite',
        'imaginary',
        'integer',
        'irrational',
        'negative',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of integer
    'integer': {
        'algebraic',
        'commutative',
        'complex',
        'composite',
        'even',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'irrational',
        'noninteger',
        'odd',
        'prime',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of irrational
    'irrational': {
        'commutative',
        'complex',
        'composite',
        'even',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'integer',
        'odd',
        'prime',
        'rational',
        'real',
        'zero',
    },

    # facts that could determine the value of negative
    'negative': {
        'commutative',
        'complex',
        'composite',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'positive',
        'prime',
        'real',
        'zero',
    },

    # facts that could determine the value of noninteger
    'noninteger': {
        'commutative',
        'composite',
        'even',
        'extended_real',
        'imaginary',
        'integer',
        'irrational',
        'odd',
        'prime',
        'zero',
    },

    # facts that could determine the value of nonnegative
    'nonnegative': {
        'commutative',
        'complex',
        'composite',
        'extended_negative',
        'extended_nonnegative',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'negative',
        'positive',
        'prime',
        'real',
        'zero',
    },

    # facts that could determine the value of nonpositive
    'nonpositive': {
        'commutative',
        'complex',
        'composite',
        'extended_nonpositive',
        'extended_positive',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'negative',
        'positive',
        'prime',
        'real',
        'zero',
    },

    # facts that could determine the value of nonzero
    'nonzero': {
        'commutative',
        'complex',
        'composite',
        'extended_nonzero',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'irrational',
        'negative',
        'odd',
        'positive',
        'prime',
        'real',
        'zero',
    },

    # facts that could determine the value of odd
    'odd': {
        'algebraic',
        'commutative',
        'complex',
        'even',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'noninteger',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of positive
    'positive': {
        'commutative',
        'complex',
        'composite',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'negative',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'prime',
        'real',
        'zero',
    },

    # facts that could determine the value of prime
    'prime': {
        'algebraic',
        'commutative',
        'complex',
        'composite',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'negative',
        'noninteger',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'positive',
        'rational',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of rational
    'rational': {
        'algebraic',
        'commutative',
        'complex',
        'composite',
        'even',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'odd',
        'prime',
        'real',
        'transcendental',
        'zero',
    },

    # facts that could determine the value of real
    'real': {
        'commutative',
        'complex',
        'composite',
        'even',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'negative',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'zero',
    },

    # facts that could determine the value of transcendental
    'transcendental': {
        'algebraic',
        'commutative',
        'complex',
        'composite',
        'even',
        'finite',
        'infinite',
        'integer',
        'odd',
        'prime',
        'rational',
        'zero',
    },

    # facts that could determine the value of zero
    'zero': {
        'algebraic',
        'commutative',
        'complex',
        'composite',
        'even',
        'extended_negative',
        'extended_nonnegative',
        'extended_nonpositive',
        'extended_nonzero',
        'extended_positive',
        'extended_real',
        'finite',
        'hermitian',
        'imaginary',
        'infinite',
        'integer',
        'irrational',
        'negative',
        'noninteger',
        'nonnegative',
        'nonpositive',
        'nonzero',
        'odd',
        'positive',
        'prime',
        'rational',
        'real',
        'transcendental',
    },

} # prereq


# Note: the order of the beta rules is used in the beta_triggers
beta_rules = [

    # Rules implying composite = True
    ({('even', True), ('positive', True), ('prime', False)},
        ('composite', True)),

    # Rules implying even = False
    ({('composite', False), ('positive', True), ('prime', False)},
        ('even', False)),

    # Rules implying even = True
    ({('integer', True), ('odd', False)},
        ('even', True)),

    # Rules implying extended_negative = True
    ({('extended_positive', False), ('extended_real', True), ('zero', False)},
        ('extended_negative', True)),
    ({('extended_nonpositive', True), ('extended_nonzero', True)},
        ('extended_negative', True)),

    # Rules implying extended_nonnegative = True
    ({('extended_negative', False), ('extended_real', True)},
        ('extended_nonnegative', True)),

    # Rules implying extended_nonpositive = True
    ({('extended_positive', False), ('extended_real', True)},
        ('extended_nonpositive', True)),

    # Rules implying extended_nonzero = True
    ({('extended_real', True), ('zero', False)},
        ('extended_nonzero', True)),

    # Rules implying extended_positive = True
    ({('extended_negative', False), ('extended_real', True), ('zero', False)},
        ('extended_positive', True)),
    ({('extended_nonnegative', True), ('extended_nonzero', True)},
        ('extended_positive', True)),

    # Rules implying extended_real = False
    ({('infinite', False), ('real', False)},
        ('extended_real', False)),
    ({('extended_negative', False), ('extended_positive', False), ('zero', False)},
        ('extended_real', False)),

    # Rules implying infinite = True
    ({('extended_real', True), ('real', False)},
        ('infinite', True)),

    # Rules implying irrational = True
    ({('rational', False), ('real', True)},
        ('irrational', True)),

    # Rules implying negative = True
    ({('positive', False), ('real', True), ('zero', False)},
        ('negative', True)),
    ({('nonpositive', True), ('nonzero', True)},
        ('negative', True)),
    ({('extended_negative', True), ('finite', True)},
        ('negative', True)),

    # Rules implying noninteger = True
    ({('extended_real', True), ('integer', False)},
        ('noninteger', True)),

    # Rules implying nonnegative = True
    ({('negative', False), ('real', True)},
        ('nonnegative', True)),
    ({('extended_nonnegative', True), ('finite', True)},
        ('nonnegative', True)),

    # Rules implying nonpositive = True
    ({('positive', False), ('real', True)},
        ('nonpositive', True)),
    ({('extended_nonpositive', True), ('finite', True)},
        ('nonpositive', True)),

    # Rules implying nonzero = True
    ({('extended_nonzero', True), ('finite', True)},
        ('nonzero', True)),

    # Rules implying odd = True
    ({('even', False), ('integer', True)},
        ('odd', True)),

    # Rules implying positive = False
    ({('composite', False), ('even', True), ('prime', False)},
        ('positive', False)),

    # Rules implying positive = True
    ({('negative', False), ('real', True), ('zero', False)},
        ('positive', True)),
    ({('nonnegative', True), ('nonzero', True)},
        ('positive', True)),
    ({('extended_positive', True), ('finite', True)},
        ('positive', True)),

    # Rules implying prime = True
    ({('composite', False), ('even', True), ('positive', True)},
        ('prime', True)),

    # Rules implying real = False
    ({('negative', False), ('positive', False), ('zero', False)},
        ('real', False)),

    # Rules implying real = True
    ({('extended_real', True), ('infinite', False)},
        ('real', True)),
    ({('extended_real', True), ('finite', True)},
        ('real', True)),

    # Rules implying transcendental = True
    ({('algebraic', False), ('complex', True)},
        ('transcendental', True)),

    # Rules implying zero = True
    ({('extended_negative', False), ('extended_positive', False), ('extended_real', True)},
        ('zero', True)),
    ({('negative', False), ('positive', False), ('real', True)},
        ('zero', True)),
    ({('extended_nonnegative', True), ('extended_nonpositive', True)},
        ('zero', True)),
    ({('nonnegative', True), ('nonpositive', True)},
        ('zero', True)),

] # beta_rules
beta_triggers = {
    ('algebraic', False): [32, 11, 3, 8, 29, 14, 25, 13, 17, 7],
    ('algebraic', True): [10, 30, 31, 27, 16, 21, 19, 22],
    ('antihermitian', False): [],
    ('commutative', False): [],
    ('complex', False): [10, 12, 11, 3, 8, 17, 7],
    ('complex', True): [32, 10, 30, 31, 27, 16, 21, 19, 22],
    ('composite', False): [1, 28, 24],
    ('composite', True): [23, 2],
    ('even', False): [23, 11, 3, 8, 29, 14, 25, 7],
    ('even', True): [3, 33, 8, 6, 5, 14, 34, 25, 20, 18, 27, 16, 21, 19, 22, 0, 28, 24, 7],
    ('extended_negative', False): [11, 33, 8, 5, 29, 34, 25, 18],
    ('extended_negative', True): [30, 12, 31, 29, 14, 20, 16, 21, 22, 17],
    ('extended_nonnegative', False): [11, 3, 6, 29, 14, 20, 7],
    ('extended_nonnegative', True): [30, 12, 31, 33, 8, 9, 6, 29, 34, 25, 18, 19, 35, 17, 7],
    ('extended_nonpositive', False): [11, 8, 5, 29, 25, 18, 7],
    ('extended_nonpositive', True): [30, 12, 31, 3, 33, 4, 5, 29, 14, 34, 20, 21, 35, 17, 7],
    ('extended_nonzero', False): [11, 33, 6, 5, 29, 34, 20, 18],
    ('extended_nonzero', True): [30, 12, 31, 3, 8, 4, 9, 6, 5, 29, 14, 25, 22, 17],
    ('extended_positive', False): [11, 3, 33, 6, 29, 14, 34, 20],
    ('extended_positive', True): [30, 12, 31, 29, 25, 18, 27, 19, 22, 17],
    ('extended_real', False): [],
    ('extended_real', True): [30, 12, 31, 3, 33, 8, 6, 5, 17, 7],
    ('finite', False): [11, 3, 8, 17, 7],
    ('finite', True): [10, 30, 31, 27, 16, 21, 19, 22],
    ('hermitian', False): [10, 12, 11, 3, 8, 17, 7],
    ('imaginary', True): [32],
    ('infinite', False): [10, 30, 31, 27, 16, 21, 19, 22],
    ('infinite', True): [11, 3, 8, 17, 7],
    ('integer', False): [11, 3, 8, 29, 14, 25, 17, 7],
    ('integer', True): [23, 2, 3, 33, 8, 6, 5, 14, 34, 25, 20, 18, 27, 16, 21, 19, 22, 7],
    ('irrational', True): [32, 3, 8, 4, 9, 6, 5, 14, 25, 15, 26, 20, 18, 27, 16, 21, 19],
    ('negative', False): [29, 34, 25, 18],
    ('negative', True): [32, 13, 17],
    ('noninteger', True): [30, 12, 31, 3, 8, 4, 9, 6, 5, 29, 14, 25, 22],
    ('nonnegative', False): [11, 3, 8, 29, 14, 20, 7],
    ('nonnegative', True): [32, 33, 8, 9, 6, 34, 25, 26, 20, 27, 21, 22, 35, 36, 13, 17, 7],
    ('nonpositive', False): [11, 3, 8, 29, 25, 18, 7],
    ('nonpositive', True): [32, 3, 33, 4, 5, 14, 34, 15, 18, 16, 19, 22, 35, 36, 13, 17, 7],
    ('nonzero', False): [29, 34, 20, 18],
    ('nonzero', True): [32, 3, 8, 4, 9, 6, 5, 14, 25, 15, 26, 20, 18, 27, 16, 21, 19, 13, 17],
    ('odd', False): [2],
    ('odd', True): [3, 8, 4, 9, 6, 5, 14, 25, 15, 26, 20, 18, 27, 16, 21, 19],
    ('positive', False): [29, 14, 34, 20],
    ('positive', True): [32, 0, 1, 28, 13, 17],
    ('prime', False): [0, 1, 24],
    ('prime', True): [23, 2],
    ('rational', False): [11, 3, 8, 29, 14, 25, 13, 17, 7],
    ('rational', True): [3, 33, 8, 6, 5, 14, 34, 25, 20, 18, 27, 16, 21, 19, 22, 17, 7],
    ('real', False): [10, 12, 11, 3, 8, 17, 7],
    ('real', True): [32, 3, 33, 8, 6, 5, 14, 34, 25, 20, 18, 27, 16, 21, 19, 22, 13, 17, 7],
    ('transcendental', True): [10, 30, 31, 11, 3, 8, 29, 14, 25, 27, 16, 21, 19, 22, 13, 17, 7],
    ('zero', False): [11, 3, 8, 29, 14, 25, 7],
    ('zero', True): [],
} # beta_triggers


generated_assumptions = {'defined_facts': defined_facts, 'full_implications': full_implications,
               'prereq': prereq, 'beta_rules': beta_rules, 'beta_triggers': beta_triggers}
